"""SMART module interfacing with PETSc SNES solver"""
import logging
from typing import Dict, List, Optional

import dolfin as d
import petsc4py.PETSc as p

from .common import Stopwatch
from .model_assembly import Compartment

logger = logging.getLogger(__name__)

__all__ = ["smartSNESProblem"]

_stopwatch_keys = [
    "snes jacobian assemble",
    "snes residual assemble",
    "snes initialize zero matrices",
]


class smartSNESProblem:
    """
    Interface to PETSc SNES solver, defining problem to be solved.

    Args:
        u: The function containing the unknown dofs (is a function from a
            `dolfin.MixedFunctionSpace`)
        Fforms: Nested list of forms for the residual ``F``. Number of
            block rows in the rhs vector is given by the number of items in the outermost list.
        Jforms_all: Nested list of forms for the Jacobian.
            Number of entries in the outermost list should be ``len(Fforms)**2``
            flattened such that ``J[i,j]=sum(Jforms_all[i*len(Fforms)+j])``,
            ``i,j=0,..,len(Fforms)-1``.
            The k-th entry of ``Jforms_all`` is a list of forms that are summed up in a given block.
        active_compartments: List of compartments used in the variational form.
        all_compartments: List of all compartments in model.
        stopwatches: Dictionary of stop-watches (stopwatch_name: stopwatch-class).
            Must include at least `snes jacobian assemble`, `snes residual assemble`
            and `snes initialize zero matrices`
        verbose: If True, output logger info.
    """

    def __init__(
        self,
        u: d.Function,
        Fforms: List[List[d.Form]],
        Jforms_all: List[List[d.Form]],
        active_compartments: List[Compartment],
        all_compartments: List[Compartment],
        stopwatches: Dict[str, Stopwatch],
        verbose: bool,
    ):
        self.u = u
        self.Fforms = Fforms
        self.Jforms_all = Jforms_all
        # for convenience, the mixed function space (model.V)
        self.W = [usub.function_space() for usub in u._functions]
        self.dim = len(self.Fforms)

        assert len(self.Jforms_all) == self.dim**2

        # Extract MPI communicator from one of the underlying meshes
        assert len(self.u._functions) >= 1
        self.comm = self.u._functions[0].function_space().mesh().mpi_comm()
        self.rank = self.comm.rank
        self.size = self.comm.size

        # save sparsity patterns of block matrices
        self.tensors = [[None] * len(Jij_list) for Jij_list in self.Jforms_all]

        # Get local_to_global maps (len=number of owned dofs + ghost dofs) and
        # dofs (len=number of owned dofs)
        self.dofs = [V.dofmap().dofs for V in self.W]
        self.lgmaps = [V.dofmap().tabulate_local_to_global_dofs().astype("int32") for V in self.W]
        self.block_sizes = [max(V.num_sub_spaces(), 1) for V in self.W]

        self.block_indices = [
            (z[::block_size] / block_size).astype("int32")
            for z, block_size in zip(self.lgmaps, self.block_sizes)
        ]
        self.lgmaps_petsc = [
            p.LGMap().create(lgmap, bsize=bsize, comm=self.comm)
            for bsize, lgmap in zip(self.block_sizes, self.lgmaps)
        ]

        self.local_ownership_ranges = [V.dofmap().ownership_range() for V in self.W]
        self.local_sizes = [x[1] - x[0] for x in self.local_ownership_ranges]
        self.global_sizes = [V.dim() for V in self.W]

        # Need sizes because some forms may be empty
        self.is_single_domain = len(self.global_sizes) == 1
        self.active_compartment_names = [c.name for c in active_compartments]
        self.mesh_id_to_name = {c.mesh_id: c.name for c in all_compartments}

        # Should we print assembly info (can get very verbose)
        self.verbose = verbose

        # Timings
        for key in _stopwatch_keys:
            if key not in stopwatches.keys():
                raise ValueError(f"Stopwatch dictionary missing stopwatch {key}")
        self.stopwatches = stopwatches

        # Check for empty forms
        self.empty_forms = []
        for i in range(self.dim):
            for j in range(self.dim):
                ij = i * self.dim + j
                if all(
                    self.Jforms_all[ij][k].function_space(0) is None
                    for k in range(len(self.Jforms_all[ij]))
                ):
                    self.empty_forms.append((i, j))
        if len(self.empty_forms) > 0:
            if self.verbose:
                logger.debug(
                    f"Forms {self.empty_forms} are empty. Skipping assembly.",
                    extra=dict(format_type="data"),
                )

    def init_petsc_matnest(self):
        Jforms = self.Jforms_all
        dim = self.dim
        Jpetsc = []
        for i in range(dim):
            for j in range(dim):
                ij = i * dim + j

                non_empty_forms = 0
                for k in range(len(Jforms[ij])):
                    if Jforms[ij][k].function_space(0) is None:
                        # The only reason this is empty is because the whole form is empty
                        assert len(Jforms[ij]) == 1
                        if self.verbose:
                            logger.debug(
                                f"{self.Jijk_name(i,j,k=None)} is empty",
                                extra=dict(format_type="log"),
                            )
                        continue
                    else:
                        non_empty_forms += 1

                    # initialize the tensor
                    if self.tensors[ij][k] is None:
                        self.tensors[ij][k] = d.PETScMatrix(self.comm)

                    logger.debug(
                        f"cpu {self.rank}: (ijk)={(i,j,k)} "
                        f"({self.local_sizes[i]}, {self.local_sizes[j]}, "
                        f"{self.global_sizes[i]}, {self.global_sizes[j]})",
                        extra=dict(format_type="log"),
                    )

                    d.assemble_mixed(
                        Jforms[ij][k], tensor=self.tensors[ij][k]
                    )  # causes issues when running multiple meshes in parallel

                if non_empty_forms == 0:
                    # If all forms are empty, we don't need to assemble. Initialize to zero matrix
                    if self.verbose:
                        logger.debug(
                            f"{self.Jijk_name(i,j)} is empty - initializing as "
                            f"empty PETSc Matrix with local size {self.local_sizes[i]}, "
                            f"{self.local_sizes[j]} "
                            f"and global size {self.global_sizes[i]}, {self.global_sizes[j]}",
                            extra=dict(format_type="log"),
                        )
                    self.tensors[ij][0] = d.PETScMatrix(self.init_petsc_matrix(i, j, assemble=True))
                    Jpetsc.append(self.tensors[ij][0])
                elif non_empty_forms == 1:
                    Jpetsc.append(self.tensors[ij][0])
                else:
                    # sum the matrices
                    # Because of the nature of these problems, it is a very reasonable
                    # guess that the matrix with the most non-zeros
                    # has a non-zero pattern which is a super-set of the other matrices.
                    nnzs = [M.nnz() for M in self.tensors[ij]]
                    k_max_nnz = nnzs.index(max(nnzs))
                    Jsum = self.d_to_p(self.tensors[ij][k_max_nnz].copy())
                    # Jsum = self.tensors[ij][k_max_nnz].copy()
                    for k in range(len(self.tensors[ij])):
                        if k == k_max_nnz:
                            continue
                        # structure options: SAME_NONZERO_PATTERN, DIFFERENT_NONZERO_PATTERN,
                        # SUBSET_NONZERO_PATTERN, UNKNOWN_NONZERO_PATTERN
                        Jsum.axpy(
                            1,
                            self.d_to_p(self.tensors[ij][k]),
                            structure=Jsum.Structure.SUBSET_NONZERO_PATTERN,
                        )
                    Jpetsc.append(d.PETScMatrix(Jsum))

        if self.is_single_domain:
            # We can't use a nest matrix
            self.Jpetsc_nest = Jpetsc[0].mat()
        else:
            self.Jpetsc_nest = d.PETScNestMatrix(Jpetsc).mat()
            # self.Jpetsc_nest = self.d_to_p(d.PETScNestMatrix(Jpetsc))
        self.Jpetsc_nest.assemble()
        logger.info(f"Jpetsc_nest assembled, size = {self.Jpetsc_nest.size}")

    def d_to_p(self, dolfin_matrix):
        return d.as_backend_type(dolfin_matrix).mat()

    def init_petsc_vecnest(self):
        dim = self.dim
        if self.verbose:
            logger.info("Initializing block residual vector", extra=dict(format_type="assembly"))

        Fpetsc = []
        for j in range(dim):
            Fsum = None
            for k in range(len(self.Fforms[j])):
                if self.Fforms[j][k].function_space(0) is None:
                    if self.verbose:
                        logger.warning(
                            f"{self.Fjk_name(j,k)}] has no function space",
                            extra=dict(format_type="log"),
                        )
                    continue

                tensor = d.PETScVector(self.comm)

                if Fsum is None:
                    Fsum = d.assemble_mixed(self.Fforms[j][k], tensor=tensor)
                else:
                    # Fsum.axpy(1, d.assemble_mixed(self.Fforms[j][k], tensor=tensor).vec(),
                    # structure=Fsum.Structure.DIFFERENT_NONZERO_PATTERN)
                    Fsum += d.assemble_mixed(self.Fforms[j][k], tensor=tensor)

            if Fsum is None:
                if self.verbose:
                    logger.debug(
                        f"{self.Fjk_name(j)} is empty - initializing as empty PETSc "
                        f"Vector with local size {self.local_sizes[j]} "
                        f"and global size {self.global_sizes[j]}",
                        extra=dict(format_type="log"),
                    )
                Fsum = d.PETScVector(self.init_petsc_vector(j, assemble=True))

            Fpetsc.append(Fsum.vec())

        if self.is_single_domain:
            # We can't use a nest vector
            self.Fpetsc_nest = d.PETScVector(Fpetsc[0]).vec()
        else:
            self.Fpetsc_nest = p.Vec().createNest(Fpetsc, comm=self.comm)
        self.Fpetsc_nest.assemble()

    def assemble_Jnest(self, Jnest):
        """Assemble Jacobian nest matrix.

        Args:
            Jnest : petsc4py.Mat
                PETSc nest matrix representing the Jacobian


        """
        if self.verbose:
            logger.debug("Assembling block Jacobian", extra=dict(format_type="assembly"))
        self.stopwatches["snes jacobian assemble"].start()
        dim = self.dim

        Jform = self.Jforms_all

        # Get the petsc sub matrices, convert to dolfin wrapper, assemble forms using
        # dolfin wrapper as tensor
        for i in range(dim):
            for j in range(dim):
                if (i, j) in self.empty_forms:
                    continue
                ij = i * dim + j
                num_subforms = len(Jform[ij])

                # Extract petsc submatrix
                if self.is_single_domain:
                    Jij_petsc = Jnest
                else:
                    Jij_petsc = Jnest.getNestSubMatrix(i, j)
                Jij_petsc.zeroEntries()  # this maintains sparse (non-zeros) structure

                if self.verbose:
                    logger.debug(
                        f"Assembling {self.Jijk_name(i,j)}:",
                        extra=dict(format_type="assembly_sub"),
                    )

                Jmats = []
                # Jijk == dFi/duj(Omega_k)
                for k in range(num_subforms):
                    # Check for empty form
                    if Jform[ij][k].function_space(0) is None:
                        if self.verbose:
                            logger.debug(
                                f"{self.Jijk_name(i,j,k)} is empty. Skipping assembly.",
                                extra=dict(format_type="data"),
                            )
                        continue

                    # if we have the sparsity pattern re-use it, if not save it for next time
                    # single domain can't re-use the tensor for some reason
                    if self.tensors[ij][k] is None and not self.is_single_domain:
                        raise AssertionError("I dont think this should happpen")
                    elif self.is_single_domain:
                        self.tensors[ij][k] = d.PETScMatrix(self.comm)
                    else:
                        if self.verbose:
                            logger.debug(
                                f"Reusing tensor for {self.Jijk_name(i,j,k)}",
                                extra=dict(format_type="data"),
                            )
                    # Assemble and append to the list of subforms
                    Jmats.append(d.assemble_mixed(Jform[ij][k], tensor=self.tensors[ij][k]))
                    # Print some useful info on assembled Jijk
                    self.print_Jijk_info(i, j, k, tensor=self.tensors[ij][k].mat())

                # Sum the assembled forms
                for Jmat in Jmats:
                    # structure options: SAME_NONZERO_PATTERN, DIFFERENT_NONZERO_PATTERN,
                    # SUBSET_NONZERO_PATTERN, UNKNOWN_NONZERO_PATTERN
                    Jij_petsc.axpy(
                        1,
                        self.d_to_p(Jmat),
                        structure=Jij_petsc.Structure.SUBSET_NONZERO_PATTERN,
                    )

                self.print_Jijk_info(i, j, k=None, tensor=Jij_petsc)

        Jnest.assemble()

        self.stopwatches["snes jacobian assemble"].pause()

    def assemble_Fnest(self, Fnest):
        """
        Assemble residual nest vector

        Arguments:
            Fnest : petsc4py.Vec
                PETSc nest vector representing the residual
        """
        dim = self.dim
        if self.verbose:
            logger.debug("Assembling block residual vector", extra=dict(format_type="assembly"))
        self.stopwatches["snes residual assemble"].start()

        if self.is_single_domain:
            Fj_petsc = [Fnest]
        else:
            Fj_petsc = Fnest.getNestSubVecs()
        Fvecs = []

        for j in range(dim):
            Fvecs.append([])
            for k in range(len(self.Fforms[j])):
                # , tensor=d.PETScVector(Fvecs[idx]))
                Fvecs[j].append(d.as_backend_type(d.assemble_mixed(self.Fforms[j][k])))
            # TODO: could probably speed this up by not using axpy if there is only one subform
            # sum the vectors
            Fj_petsc[j].zeroEntries()
            for k in range(len(self.Fforms[j])):
                Fj_petsc[j].axpy(1, Fvecs[j][k].vec())

        Fnest.assemble()
        self.stopwatches["snes residual assemble"].pause()

    def copy_u(self, unest):
        if self.is_single_domain:
            uvecs = [unest]
        else:
            uvecs = unest.getNestSubVecs()

        for idx, uvec in enumerate(uvecs):
            uvec.copy(self.u.sub(idx).vector().vec())
            self.u.sub(idx).vector().apply("")

    def F(self, snes, u, Fnest):
        self.copy_u(u)
        self.assemble_Fnest(Fnest)

    def J(self, snes, u, Jnest, P):
        self.copy_u(u)
        self.assemble_Jnest(Jnest)

    def init_petsc_matrix(self, i, j, nnz_guess=None, set_lgmap=False, assemble=False):
        """
        Initialize a PETSc matrix with appropriate structure

        Args:
            i,j : indices of the block
            nnz_guess : number of non-zeros (per row) to guess for the matrix
            assemble : whether to assemble the matrix or not (Boolean)
        """
        self.stopwatches["snes initialize zero matrices"].start()

        M = p.Mat().create(comm=self.comm)
        # ((local_nrows, global_nrows), (local_ncols, global_ncols))
        M.setSizes(
            (
                (self.local_sizes[i], self.global_sizes[i]),
                (self.local_sizes[j], self.global_sizes[j]),
            )
        )
        # M.setBlockSizes(self.block_sizes[i], self.block_sizes[j])
        # M.setBlockSizes(1,1) # seems to be ok with block size 1?
        M.setType("aij")  # "baij"

        if nnz_guess is not None:
            M.setPreallocationNNZ([nnz_guess, nnz_guess])

        # just slow down quietly if preallocation is insufficient
        # M.setOption(p.Mat.Option.NEW_NONZERO_ALLOCATION_ERR, False)

        M.setUp()

        if set_lgmap:
            M.setLGMap(self.lgmaps_petsc[i], self.lgmaps_petsc[j])

        if assemble:
            M.assemble()
        self.stopwatches["snes initialize zero matrices"].pause()

        return M

    def init_petsc_vector(self, j, assemble=False):
        """Initialize a dolfin wrapped PETSc vector with appropriate structure

        Args:
            j : index
            assemble : whether to assemble the vector or not (Boolean)
        """
        V = p.Vec().create(comm=self.comm)
        V.setSizes((self.local_sizes[j], self.global_sizes[j]))
        V.setUp()
        # V.setLGMap(self.lgmaps_petsc[j])

        if assemble:
            V.assemble()
        return V

    def Jijk_name(self, i: int, j: int, k: Optional[int] = None):
        """Get a string representation of an entry of the Jacobian.

        Args:
            i: Row index
            j: Column index
            k: If the Jacobian entry is a sum of forms, get the name
                of the domain in the k'th entry.
        """
        if k is None:
            return (
                f"J{i}{j} = dF[{self.active_compartment_names[i]}]"
                f"/du[{self.active_compartment_names[j]}]"
            )
        else:
            ij = i * self.dim + j
            if (num_entries := len(self.Jforms_all[ij])) <= k:
                raise RuntimeError(f"J[{i},{j}] only consists of {num_entries} componets")

            domain_name = self.mesh_id_to_name[self.Jforms_all[ij][k].function_space(0).mesh().id()]
            return (
                f"J{i}{j}{k} = dF[{self.active_compartment_names[i]}]"
                f"/du[{self.active_compartment_names[j]}] (domain={domain_name})"
            )

    def Fjk_name(self, j: int, k: Optional[int] = None):
        """
        Get a string representation of an entry of the residual.

        Args:
            j: Block index
            k: If the residual entry is a sum of forms, get the
                name of the domain in the k'th entry.
        """

        if k is None:
            return f"F{j} = F[{self.active_compartment_names[j]}]"
        else:
            if (num_entries := len(self.Fforms[j])) <= k:
                raise RuntimeError(f"F[{j}] only consists of {num_entries} componets")
            domain_name = self.mesh_id_to_name[self.Fforms[j][k].function_space(0).mesh().id()]
            return f"F{j} = F[{self.active_compartment_names[j]}] (domain={domain_name})"

    def print_Jijk_info(self, i, j, k=None, tensor=None):
        "Print information on Jacobian nest matrix"
        if not self.verbose:
            return
        if tensor is None:
            return
        # Print some useful info on Jijk
        info = tensor.getInfo()
        # , block_size={int(info['block_size'])}
        info_str = (
            f"size={str(tensor.size)[1:-1]: <18}, nnz={int(info['nz_allocated']): <8}, "
            f"memory[MB]={int(1e-6*info['memory']): <6}, "
            f"assemblies={int(info['assemblies']): <4}, "
            f"mallocs={int(info['mallocs']): <4}\n"
        )
        if k is None:
            logger.debug(
                f"Assembled form {self.Jijk_name(i,j,k)}:\n{info_str}",
                extra=dict(format_type="data"),
            )
        else:
            logger.debug(
                f"Assembled subform {self.Jijk_name(i,j,k)}:\n{info_str}",
                extra=dict(format_type="data"),
            )
        if info["nz_unneeded"] > 0:
            logger.warning(
                f"WARNING: {info['nz_unneeded']} nonzero entries are unneeded",
                extra=dict(format_type="warning"),
            )

    def get_csr_matrix(self, i, j):
        """This is a matrix that can be used to visualize
        the sparsity pattern using :code:`plt.spy()`"""
        if self.is_single_domain:
            M = self.Jpetsc_nest
        else:
            M = self.Jpetsc_nest.getNestSubMatrix(i, j)
        from scipy.sparse import csr_matrix

        return csr_matrix(M.getValuesCSR()[::-1], shape=M.size)



"""Classes for parameters, species, compartments, reactions, fluxes, and forms.
Model class contains functions to efficiently solve a system.
"""
import dataclasses
import logging
import numbers
import sys
from collections import OrderedDict as odict
from dataclasses import dataclass
from enum import Enum
from pprint import pformat
from textwrap import wrap
from typing import Any, List, Optional, Union
import warnings

import dolfin as d
import numpy as np
import pandas
import pint
import sympy as sym
import ufl
from cached_property import cached_property
from sympy import Symbol, integrate
from sympy.parsing.sympy_parser import parse_expr
from tabulate import tabulate

from . import common
from .config import global_settings as gset
from .units import quantity_to_unit, unit, unit_to_quantity

__all__ = [
    "empty_sbmodel",
    "sbmodel_from_locals",
    "Compartment",
    "Parameter",
    "Reaction",
    "Species",
    "sbmodel_from_locals",
    "ParameterType",
]

comm = d.MPI.comm_world
rank = comm.rank
size = comm.size
root = 0

logger = logging.getLogger(__name__)


# ====================================================
# ====================================================
# Base Classes
# ====================================================
# ====================================================


class ParameterType(str, Enum):
    from_file = "from_file"
    constant = "constant"
    expression = "expression"


class InvalidObjectException(Exception):
    pass


class ObjectContainer:
    """Parent class containing general methods used by all "containers" """

    def __init__(self, ObjectClass):  # df=None, Dict=None):
        self.Dict = odict()
        self._ObjectClass = ObjectClass
        self.properties_to_print = []  # properties to print

    # =====================================================================
    # ObjectContainer - General methods
    # ---------------------------------------------------------------------
    # General properties/methods to emulate some dictionary-like structure
    # + add some misc useful structure
    # =====================================================================

    @property
    def size(self):
        """Size of ObjectContainer"""
        return len(self.Dict)

    @property
    def items(self):
        """Return all items in container"""
        return self.Dict.items()

    @property
    def values(self):
        """Return all Dict values"""
        return self.Dict.values()

    def __iter__(self):
        """Iteratively return Dict values"""
        return iter(self.Dict.values())

    @property
    def keys(self):
        """Return all Dict keys"""
        return self.Dict.keys()

    @property
    def indices(self):
        return enumerate(self.keys)

    def to_dicts(self):
        """Returns a list of dicts that can be used to recreate the ObjectContainer"""
        return [obj.to_dict() for obj in self.values]

    def __getitem__(self, key):
        """syntactic sugar to allow: :code:`objcontainer[key] = objcontainer[key]`"""
        return self.Dict[key]

    def __setitem__(self, key, newvalue):
        """syntactic sugar to allow: :code:`objcontainer[key] = objcontainer[key]`"""
        self.Dict[key] = newvalue

    def add(self, *data):
        """Add data to object container"""
        if len(data) == 1:
            data = data[0]
            # Adding in the ObjectInstance directly
            if isinstance(data, self._ObjectClass):
                self[data.name] = data
            # Adding in an iterable of ObjectInstances
            else:
                # check if input is an iterable and if so add it item by item
                try:
                    iter(data)
                except Exception:
                    raise TypeError(
                        "Data being added to ObjectContainer must be either the "
                        "ObjectClass or an iterator."
                    )
                else:
                    if isinstance(data, dict):
                        for obj_name, obj in data.items():
                            self[obj_name] = obj
                    elif all([isinstance(obj, self._ObjectClass) for obj in data]):
                        for obj in data:
                            self[obj.name] = obj
                    else:
                        raise InvalidObjectException("Could not add data to ObjectContainer")
        # Adding via ObjectInstance arguments
        else:
            obj = self._ObjectClass(*data)
            self[obj.name] = obj

    def remove(self, name):
        """Remove data from object container"""
        if type(name) != str:
            raise TypeError("Argument must be the name of an object [str] to remove.")
        self.Dict.pop(name)

    def get_index(self, idx):
        """Get an element of the object container ordered dict by referencing its index"""
        return list(self.values)[idx]

    # ==============================================================================
    # ObjectContainer - Printing/data-formatting related methods
    # ==============================================================================

    def get_pandas_dataframe(
        self, properties_to_print: Optional[List[str]] = None, include_idx: bool = True
    ) -> pandas.DataFrame:
        """
        Create a `pandas.DataFrame` of all items in the class (defined through `self.items`)

        Args:
            properties_to_print: If set only the listed properties (by attribute name)
              is added to the series
            include_index: If true, add index as the first column in the data-frame.
        """

        df = pandas.DataFrame()
        if include_idx:
            if properties_to_print is not None and "idx" not in properties_to_print:
                properties_to_print.insert(0, "idx")
            for idx, (_, instance) in enumerate(self.items):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    # See https://github.com/hgrecco/pint-pandas/issues/128
                    df = pandas.concat(
                        [
                            df,
                            instance.get_pandas_series(
                                properties_to_print=properties_to_print, idx=idx
                            )
                            .to_frame()
                            .T,
                        ]
                    )
        else:
            for idx, (_, instance) in enumerate(self.items):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    # See https://github.com/hgrecco/pint-pandas/issues/128
                    df = pandas.concat(
                        [
                            df,
                            instance.get_pandas_series(properties_to_print=properties_to_print)
                            .to_frame()
                            .T,
                        ]
                    )

        return df

    def print_to_latex(
        self,
        properties_to_print=None,
        max_col_width=None,
        sig_figs=2,
        return_df=True,
    ):
        """
        Print object properties in latex format.
        Requires latex packages :code:`\siunitx` and :code:`\longtable`
        """

        df = self.get_pandas_dataframe_formatted(
            properties_to_print=properties_to_print,
            max_col_width=max_col_width,
            sig_figs=sig_figs,
        )

        # Change certain df entries to best format for display
        for col in df.columns:
            # Convert quantity objects to unit
            if isinstance(df[col][0], pint.Quantity):
                # if tablefmt=='latex':
                df[col] = df[col].apply(lambda x: f"${x:0.{sig_figs}e~Lx}$")

            if col == "idx":
                df = df.drop("idx", axis=1)

        if return_df:
            return df
        else:
            with pandas.option_context("max_colwidth", 1000):
                logger.info(df.to_latex(escape=False, longtable=True, index=False))

    def get_pandas_dataframe_formatted(
        self,
        properties_to_print: Optional[Union[str, List[str]]] = None,
        max_col_width=50,
        sig_figs=2,
    ):
        """Get formatted pandas dataframe for printing object properties."""
        # Get the pandas dataframe with the properties we want to print
        if properties_to_print is not None:
            if not isinstance(properties_to_print, list):
                properties_to_print = [properties_to_print]
        elif hasattr(self, "properties_to_print"):
            properties_to_print = self.properties_to_print
        df = self.get_pandas_dataframe(properties_to_print=properties_to_print, include_idx=False)

        if properties_to_print:
            df = df[properties_to_print]

        # add new lines to df entries (type str) that exceed max col width
        if max_col_width:
            for col in df.columns:
                if isinstance(df[col][0], str):
                    df[col] = df[col].apply(lambda x: "\n".join(wrap(x, max_col_width)))

        # remove leading underscores from df column names (used for cached properties)
        for col in df.columns:
            # if isinstance(df[col][0], str) and col[0] == '_':
            if col[0] == "_":
                df.rename(columns={col: col[1:]}, inplace=True)

        return df

    def print(
        self,
        tablefmt="fancy_grid",
        properties_to_print=None,
        filename=None,
        max_col_width=50,
        sig_figs=2,
    ):
        """Print object properties to file and/or terminal."""

        df = self.get_pandas_dataframe_formatted(
            properties_to_print=properties_to_print,
            max_col_width=max_col_width,
            sig_figs=sig_figs,
        )

        # # Change certain df entries to best format for printing
        for col in df.columns:
            # Convert quantity objects to unit
            if isinstance(df[col][0], pint.Quantity):
                # if tablefmt=='latex':
                df[col] = df[col].apply(lambda x: f"{x:0.{sig_figs}e~P}")

        # print to file
        if filename is None:
            logger.info(
                tabulate(df, headers="keys", tablefmt=tablefmt),
                extra=dict(format_type="table"),
            )
        else:
            original_stdout = sys.stdout  # Save a reference to the original standard output
            with open(filename, "w") as f:  # TODO: Add this to logging
                # Change the standard output to the file we created.
                sys.stdout = f
                print("This message will be written to a file.")
                print(tabulate(df, headers="keys", tablefmt=tablefmt))  # ,
                sys.stdout = original_stdout  # Reset the standard output to its original value

    def __str__(self):
        df = self.get_pandas_dataframe(properties_to_print=self.properties_to_print)
        df = df[self.properties_to_print]

        return tabulate(df, headers="keys", tablefmt="fancy_grid")


class ObjectInstance:
    """Parent class containing general methods used by all smart
    "objects": i.e. parameters, species, compartments, reactions, fluxes, forms
    """

    def _check_input_type_validity(self):
        "Check that the inputs have the same type (or are convertible) to the type hint."
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if field.type == Any:
                continue
            elif not isinstance(value, field.type):
                try:
                    setattr(self, field.name, field.type(value))
                except Exception:
                    raise TypeError(
                        f"Object {self.name!r} type error: the attribute {field.name!r} "
                        f"is expected to be {field.type}, got {type(value)} instead. "
                        "Conversion to the expected type was attempted but unsuccessful."
                    )

    def _convert_pint_quantity_to_unit(self):
        """Convert all attributes of the class that is a :code:`pint.Quantity`
        into a :code:`pint.Unit`."""
        # strip the magnitude and keep the units. Warn if magnitude!=1
        for name, attr in vars(self).items():
            if isinstance(attr, pint.Quantity):
                setattr(self, name, quantity_to_unit(attr))

    def _convert_pint_unit_to_quantity(self):
        """Convert all attributes of the class that is a :code:`pint.Unit`
        into a :code:`pint.Quantity`."""
        for name, attr in vars(self).items():
            if isinstance(attr, pint.Unit):
                setattr(self, name, unit_to_quantity(attr))

    def get_pandas_series(
        self, properties_to_print: Optional[List[str]] = None, idx: Optional[int] = None
    ):
        """
        Convert attributes of the class into a :code:`pandas.Series`.

        Args:
            properties_to_print: If set only the listed properties (by attribute name)
                is added to the series
            index: If set add to series
        """
        if properties_to_print is not None:
            dict_to_convert = odict({"idx": idx})
            dict_to_convert.update(
                odict(
                    [
                        (key, val)
                        for (key, val) in self.__dict__.items()
                        if key in properties_to_print
                    ]
                )
            )
        else:
            dict_to_convert = self.__dict__
        return pandas.Series(dict_to_convert, name=self.name)

    def print(self, properties_to_print=None):
        """Print properties in current object instance."""
        if rank == root:
            logger.info("Name: " + self.name)
            # if a custom list of properties to print is provided, only use those
            if properties_to_print:
                dict_to_print = dict(
                    [
                        (key, val)
                        for (key, val) in self.__dict__.items()
                        if key in properties_to_print
                    ]
                )
            else:
                dict_to_print = self.__dict__
                logger.info(pformat(dict_to_print, width=240))


# ==============================================================================
# ==============================================================================
# Classes for parameters, species, compartments, reactions, and fluxes
# ==============================================================================
# ==============================================================================


class ParameterContainer(ObjectContainer):
    def __init__(self):
        super().__init__(Parameter)

        self.properties_to_print = [
            "_quantity",
            "is_time_dependent",
            "sym_expr",
            "notes",
            "group",
        ]

    def print(
        self,
        tablefmt="fancy_grid",
        properties_to_print=None,
        filename=None,
        max_col_width=50,
    ):
        for s in self:
            s.quantity
        super().print(tablefmt, self.properties_to_print, filename, max_col_width)


@dataclass
class Parameter(ObjectInstance):
    """
    Parameter objects contain information for the various parameters involved in reactions,
    such as binding rates and dissociation constants.
    A Parameter object that is constant over time and space is initialized by calling

    .. code:: python

    param_var = Parameter(
        name, value, unit, group (opt), notes (opt),
        use_preintegration (opt)
        )

    Args:
        name: string naming the parameter
        value: value of the given parameter
        unit: units associated with given value
        group (optional): string placing this parameter in a designated group;
             for organizational purposes when there are multiple reaction modules
        notes (optional): string related to this parameter
        use_preintegration (optional): not applicable for constant parameter

    To initialize a parameter object that varies over time, you can either
    specify a string that gives the parameter as a function of time (t)
    or load data from a .txt file.

    To load from a string expression, call:

    .. code:: python

        param_var = Parameter.from_expression(
            name, sym_expr, unit, preint_sym_expr (opt), group (opt),
            notes (opt), use_preintegration (opt)
        )

    Inputs are the same as described above, except:

    Args:
        sym_expr: string specifying an expression, "t" should be the only free variable
        preint_sym_expr (optional): string giving the integral of the expression; if not given
                                  and use_preintegration is true, then sympy tries to integrate
                                  using sympy.integrate()
        use_preintegration (optional):  use preintegration in solution process if
                                     "use_preintegration" is true (defaults to false)

    To load parameter over time from a .txt file, call:

    .. code:: python

        param_var = Parameter.from_file(
            name, sampling_file, unit, group (opt),
            notes (opt), use_preintegration (opt)
        )

    Inputs are the same as described above, except:

    Args:
        sampling_file: name of text file giving parameter data in two columns (comma-separated) -
                     first column is time (starting with t=0.0) and second is parameter values
        use_preintegration (optional):  use preintegration in solution process if
                                     "use_preintegration" is true (defaults to false),
                                     uses sci.integrate.cumtrapz for numerical integration
    """

    name: str
    value: float
    unit: pint.Unit
    group: str = ""
    notes: str = ""
    use_preintegration: bool = False

    def to_dict(self):
        """Convert to a dict that can be used to recreate the object."""
        keys_to_keep = [
            "name",
            "value",
            "unit",
            "group",
            "notes",
            "use_preintegration",
            "is_time_dependent",
            "is_space_dependent",
            "sym_expr",
            "preint_sym_expr",
            "sampling_data",
            "preint_sampling_data",
            "type",
            "free_symbols",
        ]
        return {key: self.__dict__[key] for key in keys_to_keep}

    @classmethod
    def from_dict(cls, input_dict):
        """Read parameter object from Dict"""
        parameter = cls(input_dict["name"], input_dict["value"], input_dict["unit"])
        for key, val in input_dict.items():
            setattr(parameter, key, val)
        parameter.__post_init__()
        return parameter

    @classmethod
    def from_file(cls, name, sampling_file, unit, group="", notes="", use_preintegration=False):
        """ "
        Load in a purely time-dependent scalar function from data
        Data needs to be read in from a text file with two columns
        where the first column is time (first entry must be 0.0)
        and the second column is the parameter values.
        Columns should be comma-separated.
        """
        # load in sampling data file
        sampling_data = np.genfromtxt(sampling_file, dtype="float", delimiter=",")
        logger.info(f"Loading in data for parameter {name}", extra=dict(format_type="log"))
        if sampling_data[0, 0] != 0.0 or sampling_data.shape[1] != 2:
            raise NotImplementedError
        value = sampling_data[0, 1]  # initial value

        parameter = cls(
            name,
            value,
            unit,
            group=group,
            notes=notes,
            use_preintegration=use_preintegration,
        )

        if use_preintegration:
            # preintegrate sampling data using cumtrapz
            from scipy.integrate import cumtrapz

            int_data = cumtrapz(sampling_data[:, 1], x=sampling_data[:, 0], initial=0)
            # concatenate time vector
            preint_sampling_data = np.hstack(
                sampling_data[:, 0].reshape(-1, 1), int_data.reshape(-1, 1)
            )
            parameter.preint_sampling_data = preint_sampling_data

        # initialize instance
        parameter.sampling_file = sampling_file
        parameter.sampling_data = sampling_data
        parameter.is_time_dependent = True
        parameter.is_space_dependent = False  # not supported yet
        parameter.type = ParameterType.from_file
        parameter.__post_init__()
        logger.info(
            f"Time-dependent parameter {name} loaded from file.",
            extra=dict(format_type="log"),
        )

        return parameter

    @classmethod
    def from_expression(
        cls,
        name,
        sym_expr,
        unit,
        preint_sym_expr=None,
        group="",
        notes="",
        use_preintegration=False,
    ):
        """Use sympy to parse time-dependent expression for parameter"""
        # Parse the given string to create a sympy expression
        if isinstance(sym_expr, str):
            sym_expr = parse_expr(sym_expr)
        x, y, z = (Symbol(f"x[{i}]") for i in range(3))
        sym_expr = sym_expr.subs({"x": x, "y": y, "z": z})

        # Check if expression is time/space dependent
        free_symbols = [str(x) for x in sym_expr.free_symbols]
        is_time_dependent = "t" in free_symbols
        is_space_dependent = not {"x[0]", "x[1]", "x[2]"}.isdisjoint(set(free_symbols))
        if is_space_dependent:
            raise NotImplementedError
        # For now, parameters can only be defined in terms of time/space
        if not {"x[0]", "x[1]", "x[2]", "t"}.issuperset(free_symbols):
            raise NotImplementedError

        # TODO: fix this when implementing space dependent parameters
        if is_time_dependent:
            value = float(sym_expr.subs({"t": 0.0}))

        parameter = cls(
            name,
            value,
            unit,
            group=group,
            notes=notes,
            use_preintegration=use_preintegration,
        )

        if use_preintegration:
            if preint_sym_expr:
                if isinstance(preint_sym_expr, str):
                    preint_sym_expr = parse_expr(preint_sym_expr)
                preint_sym_expr = preint_sym_expr.subs({"x": x, "y": y, "z": z})
            else:
                # try to integrate
                t = Symbol("t")
                preint_sym_expr = integrate(sym_expr, t)
            parameter.preint_sym_expr = preint_sym_expr

        parameter.free_symbols = free_symbols
        parameter.sym_expr = sym_expr
        parameter.is_time_dependent = is_time_dependent
        parameter.is_space_dependent = is_space_dependent

        # parameter.dolfin_expression = d.Expression(sym.printing.ccode(sym_expr), t=0.0, degree=1)
        parameter.type = ParameterType.expression
        parameter.__post_init__()
        logger.debug(
            f"Time-dependent parameter {name} evaluated from expression.",
            extra=dict(format_type="log"),
        )

        return parameter

    def __post_init__(self):
        if not hasattr(self, "is_time_dependent"):
            self.is_time_dependent = False
        if not hasattr(self, "is_space_dependent"):
            self.is_space_dependent = False

        if self.use_preintegration:
            logger.warning(
                f"Warning! Pre-integrating parameter {self.name}. Make sure that "
                f"expressions {self.name} appears in have no other time-dependent variables.",
                extra=dict(format_type="warning"),
            )

        attributes = [
            "sym_expr",
            "preint_sym_expr",
            "sampling_file",
            "sampling_data",
            "preint_sampling_data",
            "free_symbols",
        ]
        for attribute in attributes:
            if not hasattr(self, attribute):
                setattr(self, attribute, None)

        if not hasattr(self, "type"):
            self.type = ParameterType.constant

        self._convert_pint_quantity_to_unit()
        self._check_input_type_validity()
        self._convert_pint_unit_to_quantity()
        self.check_validity()
        self.value_vector = np.array([0, self.value])

    @property
    def dolfin_quantity(self):
        if hasattr(self, "dolfin_expression") and not self.use_preintegration:
            return self.dolfin_expression * self.unit
        else:
            return self.dolfin_constant * self.unit

    @property
    def quantity(self):
        self._quantity = self.value * self.unit
        return self._quantity

    def check_validity(self):
        """Confirm that time-dependent parameter is defined in terms of time"""
        if self.is_time_dependent:
            if all(
                [x in ("", None) for x in [self.sampling_file, self.sym_expr, self.preint_sym_expr]]
            ):
                raise ValueError(
                    f"Parameter {self.name} is marked as time dependent "
                    "but is not defined in terms of time."
                )


class SpeciesContainer(ObjectContainer):
    def __init__(self):
        super().__init__(Species)
        self.properties_to_print = ["compartment_name", "dof_index", "_Diffusion"]

    def print(
        self,
        tablefmt="fancy_grid",
        properties_to_print=None,
        filename=None,
        max_col_width=50,
    ):
        for s in self:
            s.D_quantity
            s.latex_name
        super().print(tablefmt, self.properties_to_print, filename, max_col_width)

    def print_to_latex(
        self,
        properties_to_print=None,
        max_col_width=None,
        sig_figs=2,
        return_df=False,
    ):
        properties_to_print = ["_latex_name"]
        properties_to_print.extend(self.properties_to_print)
        df = super().print_to_latex(properties_to_print, max_col_width, sig_figs, return_df=True)
        # fix dof_index
        for col in df.columns:
            if col == "dof_index":
                df[col] = df[col].astype(int)
        # fix name
        # get the column of df that contains the name
        # this can be more robust
        # df.columns

        if return_df:
            return df
        else:
            with pandas.option_context("max_colwidth", 1000):
                logger.info(df.to_latex(escape=False, longtable=True, index=False))


@dataclass
class Species(ObjectInstance):
    """
    Each Species object contains information for one state variable in the model
    (can be a molecule, receptor open probability, membrane voltage, etc.)

    Args:
        name: string naming the species
        conc_init: initial concentration for this species
            (can be an expression given by a string to be parsed by sympy
            - the only unknowns in the expression should be x, y, and z)
        conc_units: concentration units for this species
        D: diffusion coefficient value
        diffusion_units: units for diffusion coefficient
        compartment_name: each species should be assigned to a single compartment
        group (optional): for larger models, specifies a group of species this belongs to;
            for organizational purposes when there are multiple reaction modules

    Species object is initialized by calling:

    .. code:: python

        species_var = Species(
            name, initial_condition, concentration_units,
            D, diffusion_units, compartment_name, group (opt)
        )
    """

    name: str
    initial_condition: Any
    concentration_units: pint.Unit
    D: float
    diffusion_units: pint.Unit
    compartment_name: str
    group: str = ""

    def to_dict(self):
        "Convert to a dict that can be used to recreate the object."
        keys_to_keep = [
            "name",
            "initial_condition",
            "concentration_units",
            "D",
            "diffusion_units",
            "compartment_name",
            "group",
        ]
        return {key: self.__dict__[key] for key in keys_to_keep}

    @classmethod
    def from_dict(cls, input_dict):
        """Load Species object from Dict"""
        return cls(**input_dict)

    def __post_init__(self):
        # self.sub_species = {} # additional compartments this species
        # may live in in addition to its primary one
        self.is_in_a_reaction = False
        self.is_an_added_species = False
        self.dof_map = None
        self.u = dict()
        self._usplit = dict()
        self.ut = None
        self.v = None

        if isinstance(self.initial_condition, float):
            pass
        elif isinstance(self.initial_condition, int):
            self.initial_condition = float(self.initial_condition)
        elif isinstance(self.initial_condition, str):
            x, y, z = (Symbol(f"x[{i}]") for i in range(3))
            # Parse the given string to create a sympy expression
            sym_expr = parse_expr(self.initial_condition).subs({"x": x, "y": y, "z": z})

            # Check if expression is space dependent
            free_symbols = [str(x) for x in sym_expr.free_symbols]
            if not {"x[0]", "x[1]", "x[2]"}.issuperset(free_symbols):
                raise NotImplementedError
            logger.debug(
                f"Creating dolfin object for space-dependent initial condition {self.name}",
                extra=dict(format_type="log"),
            )
            self.initial_condition_expression = d.Expression(sym.printing.ccode(sym_expr), degree=1)
        else:
            raise TypeError("initial_condition must be a float or string.")

        self._convert_pint_quantity_to_unit()
        self._check_input_type_validity()
        self._convert_pint_unit_to_quantity()
        self.check_validity()

    def check_validity(self):
        """Species validity checks:

        * Initial condition is greater than or equal to 0
        * Diffusion coefficient is greater than or equal to 0
        * Diffusion coefficient has units of length^2/time
        """
        # checking values
        if isinstance(self.initial_condition, float) and self.initial_condition < 0.0:
            raise ValueError(
                f"Initial condition for species {self.name} must be greater or equal to 0."
            )
        if self.D < 0.0:
            raise ValueError(
                f"Diffusion coefficient for species {self.name} must be greater or equal to 0."
            )
        # checking units
        if not self.diffusion_units.check("[length]^2/[time]"):
            raise ValueError(
                f"Units of diffusion coefficient for species {self.name} must "
                "be dimensionally equivalent to [length]^2/[time]."
            )

    @cached_property
    def vscalar(self):
        return d.TestFunction(common.sub(self.compartment.V, 0, True))

    @property
    def dolfin_quantity(self):
        return self._usplit["u"] * self.concentration_units

    @property
    def initial_condition_quantity(self):
        self._Initial_Concentration = self.initial_condition * self.concentration_units
        return self._Initial_Concentration

    @property
    def D_quantity(self):
        self._Diffusion = self.D * self.diffusion_units
        return self._Diffusion

    @property
    def sym(self):
        self._sym = Symbol(self.name)
        return self._sym

    @property
    def latex_name(self):
        # Change _ to - in name
        name = self.name.replace("_", "-")
        self._latex_name = sym.latex(Symbol(name))
        return self._latex_name


class CompartmentContainer(ObjectContainer):
    def __init__(self):
        super().__init__(Compartment)

        self.properties_to_print = [
            "_mesh_id",
            "dimensionality",
            "num_species",
            "_num_vertices",
            "_num_dofs",
            "_num_dofs_local",
            "_num_cells",
            "cell_marker",
            "_nvolume",
        ]

    def print(
        self,
        tablefmt="fancy_grid",
        properties_to_print=None,
        filename=None,
        max_col_width=50,
    ):
        for c in self:
            c.mesh_id
            c.nvolume
            c.num_vertices
            c.num_dofs
            c.num_dofs_local
            c.num_cells
        super().print(tablefmt, self.properties_to_print, filename, max_col_width)


@dataclass
class Compartment(ObjectInstance):
    """Each Compartment object contains information describing a surface, volume, or edge
    within the geometry of interest. The object is initialized by calling:

        .. code:: python

            compartment_var = Compartment(name, dimensionality, compartment_units, cell_marker)

    Args:
        name: string naming the compartment
        dimensionality: topological dimensionality (e.g. 3 for volume, 2 for surface)
        compartment_units: length units for the compartment
        cell_marker: marker value identifying the compartment in the parent mesh
    """

    name: str
    dimensionality: int
    compartment_units: pint.Unit
    cell_marker: Any

    def to_dict(self):
        "Convert to a dict that can be used to recreate the object."
        keys_to_keep = ["name", "dimensionality", "compartment_units", "cell_marker"]
        return {key: self.__dict__[key] for key in keys_to_keep}

    @classmethod
    def from_dict(cls, input_dict):
        return cls(**input_dict)

    def __post_init__(self):
        if (
            isinstance(self.cell_marker, list)
            and not all([isinstance(m, int) for m in self.cell_marker])
            or not isinstance(self.cell_marker, (int, list))
        ):
            raise TypeError("cell_marker must be an int or list of ints.")

        self._convert_pint_quantity_to_unit()
        self._check_input_type_validity()
        self._convert_pint_unit_to_quantity()
        self.check_validity()

        # Initialize
        self.species = odict()
        self.u = dict()
        self._usplit = dict()
        self.V = None
        self.v = None

    def check_validity(self):
        """
        Compartment validity checks:

        * Compartment dimensionality is 1,2, or 3
        * Compartment units are of type "length"
        """
        if self.dimensionality not in [1, 2, 3]:
            raise ValueError(
                f"Compartment {self.name} has dimensionality {self.dimensionality}. "
                "Dimensionality must be in [1,2,3]."
            )
        # checking units
        if not self.compartment_units.check("[length]"):
            raise ValueError(
                f"Compartment {self.name} has units of {self.compartment_units} "
                "- units must be dimensionally equivalent to [length]."
            )

    def specify_nonadjacency(self, nonadjacent_compartment_list=None):
        """Specify if this compartment is NOT adjacent to another compartment.
        Not necessary, but will speed-up initialization of very large problems.
        Only needs to be specified for surface meshes as those are the
        ones that MeshViews are built on.
        """
        if nonadjacent_compartment_list is None:
            self.nonadjacent_compartment_list = []
        self.nonadjacent_compartment_list = nonadjacent_compartment_list

    @property
    def measure_units(self):
        return self.compartment_units**self.dimensionality

    @property
    def mesh_id(self):
        self._mesh_id = self.mesh.id
        return self._mesh_id

    @property
    def dolfin_mesh(self):
        return self.mesh.dolfin_mesh

    @property
    def nvolume(self):
        """nvolume with proper units"""
        self._nvolume = self.mesh.nvolume * self.compartment_units**self.dimensionality
        return self._nvolume

    @property
    def num_cells(self):
        self._num_cells = self.mesh.num_cells
        return self._num_cells

    @property
    def num_facets(self):
        self._num_facets = self.mesh.num_facets
        return self._num_facets

    @property
    def num_vertices(self):
        self._num_vertices = self.mesh.num_vertices
        return self._num_vertices

    @property
    def num_dofs(self):
        """Number of degrees of freedom for this compartment"""
        # self._num_dofs = self.num_species * self.num_vertices
        # return self._num_dofs
        if self.V is None:
            self._num_dofs = 0
        else:
            self._num_dofs = self.V.dim()
        return self._num_dofs

    @property
    def num_dofs_local(self):
        """Number of degrees of freedom for this compartment, local to this process"""
        if self.V is None:
            self._num_dofs_local = 0
        else:
            self._ownership_range = self.V.dofmap().ownership_range()
            self._num_dofs_local = self._ownership_range[1] - self._ownership_range[0]
        return self._num_dofs_local


class ReactionContainer(ObjectContainer):
    def __init__(self):
        super().__init__(Reaction)

        self.properties_to_print = ["lhs", "rhs", "eqn_f_str", "eqn_r_str"]

    def print(
        self,
        tablefmt="fancy_grid",
        properties_to_print=None,
        filename=None,
        max_col_width=50,
    ):
        super().print(tablefmt, self.properties_to_print, filename, max_col_width)


@dataclass
class Reaction(ObjectInstance):
    """
    A Reaction object contains information on a single biochemical interaction
    between species in a single compartment or across multiple compartments.

    Args:
        name: string naming the reaction
        lhs: list of strings specifying the reactants for this reaction
        rhs: list of strings specifying the products for this reaction
            NOTE: the lists "lhs" and "rhs" determine the stoichiometry of the reaction;
            for instance, if two A's react to give one B, the reactants list would be
            :code:`["A","A"]`, and the products list would be :code:`["B"]`
        param_map: relationship between the parameters specified in the reaction string
            and those given in the parameter container. By default, the reaction parameters are
            "on" and "off" when a system obeys simple mass action.
            If the forward rate is given by a parameter :code:`k1` and the reverse
            rate is given by :code:`k2`, then :code:`param_map = {"on":"k1", "off":"k2"}`
        eqn_f_str: For systems not obeying simple mass action,
            this string specifies the forward reaction rate By default,
            this string is "on*{all reactants multiplied together}"
        eqn_r_str: For systems not obeying simple mass action,
            this string specifies the reverse reaction rate
            By default, this string is :code:`off*{all products multiplied together}`
            reaction_type: either "custom" or "mass_action" (default is "mass_action")
            [never a required argument]
        species_map: same format as param_map;
            required if the species does not appear in the lhs or rhs lists
        explicit_restriction_to_domain: string specifying where the reaction occurs;
            required if the reaction is not constrained by the reaction string
            (e.g., if production occurs only at the boundary,
            but the species being produced exists through the entire volume)
        group: string placing this reaction in a reaction group;
            for organizational purposes when there are multiple reaction modules
            flux_scaling: in certain cases, a given reactant or product
            may experience a scaled flux (for instance, if we assume that
            some of the molecules are immediately sequestered after the reaction);
            in this case, to signify that this flux should be rescaled, we specify
            :code:`flux_scaling = {scaled_species: scale_factor}`,
            where scaled_species is a string specifying the species to be scaled and
            scale_factor is a number specifying the rescaling factor

    .. note::

        The Reaction object is initialized by calling:

        .. code:: python

            reaction_name = Reaction(
                name, lhs, rhs, param_map,
                eqn_f_str (opt), eqn_r_str (opt), reaction_type (opt), species_map,
                explicit_restriction_to_domain (opt), group (opt), flux_scaling (opt)
            )
    """

    name: str
    lhs: list
    rhs: list
    param_map: dict
    species_map: dict = dataclasses.field(default_factory=dict)
    flux_scaling: dict = dataclasses.field(default_factory=dict)
    reaction_type: str = "mass_action"
    explicit_restriction_to_domain: str = ""
    track_value: bool = False
    eqn_f_str: str = ""
    eqn_r_str: str = ""
    group: str = ""

    def to_dict(self):
        "Convert to a dict that can be used to recreate the object."
        keys_to_keep = [
            "name",
            "lhs",
            "rhs",
            "param_map",
            "species_map",
            "flux_scaling",
            "reaction_type",
            "explicit_restriction_to_domain",
            "track_value",
            "eqn_f_str",
            "eqn_r_str",
            "group",
        ]
        return {key: self.__dict__[key] for key in keys_to_keep}

    @classmethod
    def from_dict(cls, input_dict):
        return cls(**input_dict)

    def __post_init__(self):
        self._check_input_type_validity()
        self.check_validity()
        self.fluxes = dict()

        if self.eqn_f_str != "" or self.eqn_r_str != "" and self.reaction_type == "mass_action":
            self.reaction_type = "custom"

        # Finish initializing the species map
        for species_name in set(self.lhs + self.rhs):
            if species_name not in self.species_map:
                self.species_map[species_name] = species_name

        # Finish initializing the species flux scaling
        for species_name in set(self.lhs + self.rhs):
            if species_name not in self.flux_scaling:
                self.flux_scaling[species_name] = None

    def check_validity(self):
        """
        Reaction validity checks:

        * LHS (reactants) and RHS (products) are specified as lists of strings
        * param_map must be specified as a dict of "str:str"
        * Species_map must be specified as a dict of "str:str"
        * If given, flux scaling must be specified as a dict of "species:scale_factor"
        """
        # Type checking
        if not all([isinstance(x, str) for x in self.lhs]):
            raise TypeError(f"Reaction {self.name} requires a list of strings as input for lhs.")
        if not all([isinstance(x, str) for x in self.rhs]):
            raise TypeError(f"Reaction {self.name} requires a list of strings as input for rhs.")
        if not all([type(k) == str and type(v) == str for (k, v) in self.param_map.items()]):
            raise TypeError(
                f"Reaction {self.name} requires a dict of str:str as input for param_map."
            )
        if self.species_map:
            if not all(
                [isinstance(k, str) and isinstance(v, str) for (k, v) in self.species_map.items()]
            ):
                raise TypeError(
                    f"Reaction {self.name} requires a dict of str:str as input for species_map."
                )
        if self.flux_scaling:
            if not all(
                [
                    isinstance(k, str) and isinstance(v, (numbers.Number, None))
                    for (k, v) in self.flux_scaling.items()
                ]
            ):
                raise TypeError(
                    f"Reaction {self.name} requires a dict of "
                    "str:number as input for flux_scaling."
                )

    def _parse_custom_reaction(self, reaction_eqn_str):
        "Substitute parameters and species into reaction expression"
        reaction_expr = parse_expr(reaction_eqn_str)
        reaction_expr = reaction_expr.subs(self.param_map)
        reaction_expr = reaction_expr.subs(self.species_map)
        return str(reaction_expr)

    def reaction_to_fluxes(self):
        """
        Convert reactions to fluxes -
        in general, for each product and each reactant there are two fluxes,
        one forward flux (dictated by :code:`self.eqn_f_str`)
        and one reverse flux (dictated by :code:`self.eqn_r_str`),
        stoichiometry is dictated by the number of times a given species occurs on the lhs or rhs
        """
        logger.debug(f"Getting fluxes for reaction {self.name}", extra=dict(format_type="log"))
        # set of 2-tuples. (species_name, signed stoichiometry)
        self.species_stoich = {
            (species_name, -1 * self.lhs.count(species_name)) for species_name in self.lhs
        }
        self.species_stoich.update(
            {(species_name, 1 * self.rhs.count(species_name)) for species_name in self.rhs}
        )
        # convert to dict
        self.species_stoich = dict(self.species_stoich)
        # combine with user-defined flux scaling
        for species_name in self.species_stoich.keys():
            if self.flux_scaling[species_name] is not None:
                logger.debug(
                    f"Flux {self.name}: stoichiometry/flux for species {species_name} "
                    f"scaled by {self.flux_scaling[species_name]}",
                    extra=dict(format_type="log"),
                )
                self.species_stoich[species_name] *= self.flux_scaling[species_name]

        for species_name, stoich in self.species_stoich.items():
            species = self.species[species_name]
            if self.eqn_f_str:
                flux_name = self.name + f" [{species_name} (f)]"
                eqn = stoich * parse_expr(self.eqn_f_str)
                self.fluxes.update({flux_name: Flux(flux_name, species, eqn, self)})
            if self.eqn_r_str:
                flux_name = self.name + f" [{species_name} (r)]"
                eqn = -stoich * parse_expr(self.eqn_r_str)
                self.fluxes.update({flux_name: Flux(flux_name, species, eqn, self)})


class FluxContainer(ObjectContainer):
    def __init__(self):
        super().__init__(Flux)

        self.properties_to_print = [
            "_species_name",
            "equation",
            "topology",
            "_assembled_flux",
        ]

    def print(
        self,
        tablefmt="fancy_grid",
        properties_to_print=None,
        filename=None,
        max_col_width=50,
    ):
        for f in self:
            f.assembled_flux
            f.equation_lambda_eval("quantity")
        super().print(tablefmt, self.properties_to_print, filename, max_col_width)


@dataclass
class Flux(ObjectInstance):
    """Flux objects are created from reaction objects and should not be
    explicitly initialized by the user.
    Each flux object contains:

    * name: string (created as reaction name + species name + (f) or (r))
    * destination_species: flux increases or decreases this species
    * equation: directionality * stoichiometry * reaction string
    * reaction: reaction object this flux comes from
    """

    name: str
    destination_species: Species
    equation: sym.Expr
    reaction: Reaction

    def check_validity(self):
        "No validity checks for flux objects currently"
        pass

    def __post_init__(self):
        # for nice printing
        self._species_name = self.destination_species.name

        self._check_input_type_validity()
        self.check_validity()

        # Add in an uninitialized unit_scale_factor
        self.unit_scale_factor = 1.0 * unit.dimensionless
        self.equation = self.equation * Symbol("unit_scale_factor")

        # Getting additional flux properties
        self._post_init_get_involved_species_parameters_compartments()
        self._post_init_get_flux_topology()

        # Get equation lambda expression
        self.equation_lambda = sym.lambdify(
            list(self.equation_variables.keys()),
            self.equation,
            modules=common.smart_expressions(gset["dolfin_expressions"]),
        )

        # Update equation with correct unit scale factor
        self._post_init_get_flux_units()
        self._post_init_get_integration_measure()

        # Check if flux is linear with respect to different components
        self.is_linear_wrt_comp = dict()
        self._post_init_get_is_linear_comp()

    def _post_init_get_involved_species_parameters_compartments(self):
        "Find species, parameters, and compartments involved in this flux"
        self.destination_compartment = self.destination_species.compartment

        # Get the subset of species/parameters/compartments that are relevant
        variables = {str(x) for x in self.equation.free_symbols}
        all_params = self.reaction.parameters
        all_species = self.reaction.species
        self.parameters = {x: all_params[x] for x in variables.intersection(all_params.keys())}
        self.species = {x: all_species[x] for x in variables.intersection(all_species.keys())}
        self.compartments = self.reaction.compartments

    def _post_init_get_flux_topology(self):
        """
        "Flux topology" refers to what types of compartments the flux occurs over
        For instance, if a reaction occurs only in a volume, its topology is type "volume"
        or if it involves species moving from one volume to another (across a surface),
        it is of type "volume-surface_to_volume". Depending on the flux topology,
        the flux will contribute to the system either as a term in the PDE or as
        a boundary condition.

        The dimensionality indicated in the left brackets below refers to how many
        compartments are involved in the given flux (e.g. 3d = 3 compartments)

        Flux topology types:
        [1d] volume:                    PDE of u
        [1d] surface:                   PDE of v
        [2d] volume_to_surface:         PDE of v
        [2d] surface_to_volume:         BC of u
        [3d] volume-surface_to_volume:  BC of u ()
        [3d] volume-volume_to_surface:  PDE of v ()
        """
        # 1 compartment flux
        if self.reaction.topology in ["surface", "volume"]:
            self.topology = self.reaction.topology
            source_compartments = {self.destination_compartment.name}
        # 2 or 3 compartment flux
        elif self.reaction.topology in ["volume_surface", "volume_surface_volume"]:
            source_compartments = set(self.compartments.keys()).difference(
                {self.destination_compartment.name}
            )

            if self.reaction.topology == "volume_surface":
                assert len(source_compartments) == 1
                if self.destination_compartment.is_volume:
                    self.topology = "surface_to_volume"
                else:
                    self.topology = "volume_to_surface"

            elif self.reaction.topology == "volume_surface_volume":
                assert len(source_compartments) == 2
                if self.destination_compartment.is_volume:
                    self.topology = "volume-surface_to_volume"
                else:
                    self.topology = "volume-volume_to_surface"

        else:
            raise AssertionError()

        self.source_compartments = {
            name: self.reaction.compartments[name] for name in source_compartments
        }
        self.surface = [c for c in self.compartments.values() if c.mesh.is_surface]
        if len(self.surface) == 1:
            self.surface = self.surface[0]
        else:
            self.surface = None

        # Based on topology we know if it is a boundary condition or RHS term
        if self.topology in [
            "volume",
            "surface",
            "volume_to_surface",
            "volume-volume_to_surface",
        ]:
            self.is_boundary_condition = False
        elif self.topology in ["surface_to_volume", "volume-surface_to_volume"]:
            self.is_boundary_condition = True
        else:
            raise AssertionError()

    def _post_init_get_flux_units(self):
        """
        Check that flux units match expected type of units.
        For boundary conditions, the flux units should be
        (concentration_units / compartment_units) * diffusion_units
        For fluxes that contribute to the PDE terms, units should be
        (concentration_units / time)
        If the units dimensionality does not match the expected form, an error is thrown;
        if the dimensionality matches, but the units scaling is not 1.0
        (e.g. if we have a flux specified as nM/s, but concentration_units are defined as uM),
        then self.unit_scale_factor is set to the appropriate factor
        (1000 in the example where we need to convert nM -> uM)
        NOTE: All unit checks are completed using pint, which may not be compatible with
        certain functions such as sign().
        """
        concentration_units = self.destination_species.concentration_units
        compartment_units = self.destination_compartment.compartment_units
        diffusion_units = self.destination_species.diffusion_units

        # The expected units
        if self.is_boundary_condition:
            self._expected_flux_units = (
                1.0 * concentration_units / compartment_units * diffusion_units
            )  # ~D*du/dn
        else:
            self._expected_flux_units = 1.0 * concentration_units / unit.s  # rhs term. ~du/dt

        # Use the uninitialized unit_scale_factor to get the actual units
        # this is redundant if called by __post_init__
        self.unit_scale_factor = 1.0 * unit.dimensionless
        initial_equation_units = self.equation_lambda_eval("units")

        # If unit dimensionality is not correct a parameter likely needs to be adjusted
        if self._expected_flux_units.dimensionality != initial_equation_units.dimensionality:
            logger.info(self.unit_scale_factor)
            raise ValueError(
                f"Flux {self.name} has wrong units (cannot be converted) "
                f"- expected {self._expected_flux_units}, got {initial_equation_units}."
            )
        # Fix scaling
        else:
            # Define new unit_scale_factor, and update equation_units
            # by re-evaluating the lambda expression
            self.unit_scale_factor = (
                initial_equation_units.to(self._expected_flux_units) / initial_equation_units
            )
            self.equation_units = self.equation_lambda_eval(
                "units"
            )  # these should now be the proper units

            # should be redundant with previous checks, but just in case
            assert self.unit_scale_factor.dimensionless
            assert (
                initial_equation_units * self.unit_scale_factor
            ).units == self._expected_flux_units
            assert self.equation_units == self._expected_flux_units

            # If we already have the correct units, there is no need to update the equation
            if self.unit_scale_factor.magnitude == 1.0:
                return

            logger.debug(
                f"Flux {self.name} scaled by {self.unit_scale_factor}",
                extra=dict(new_lines=[1, 0], format_type="log"),
            )
            logger.debug(f"Old flux units: {self.equation_units}", extra=dict(format_type="log"))
            logger.debug(
                f"New flux units: {self._expected_flux_units}",
                extra=dict(new_lines=[0, 1], format_type="log"),
            )
            print("")

    def _post_init_get_integration_measure(self):
        """
        Flux topologies (cf. definitions in _post_init_get_flux_topology above):
        [1d] volume:                    PDE of u
        [1d] surface:                   PDE of v
        [2d] volume_to_surface:         PDE of v
        [2d] surface_to_volume:         BC of u
        [3d] volume-surface_to_volume:  BC of u ()
        [3d] volume-volume_to_surface:  PDE of v ()

        1d means no other compartment are involved, so the integration
        measure is the volume of the compartment
        2d/3d means two/three compartments are involved, so the
        integration measure is the intersection between all compartments
        """

        if self.topology in ["volume", "surface"]:
            self.measure = self.destination_compartment.mesh.dx
            self.measure_units = self.destination_compartment.measure_units
        elif self.topology in [
            "volume_to_surface",
            "surface_to_volume",
            "volume-volume_to_surface",
            "volume-surface_to_volume",
        ]:
            # intersection of this surface with boundary of source volume(s)
            logger.debug(
                "DEBUGGING INTEGRATION MEASURE (only fully defined domains are enabled for now)"
            )
            self.measure = self.surface.mesh.dx
            self.measure_units = self.surface.compartment_units**self.surface.dimensionality

    # We define this as a property so that it is automatically updated

    @property
    def equation_variables(self):
        variables = {
            variable.name: variable.dolfin_quantity
            for variable in {**self.parameters, **self.species}.values()
        }
        variables.update({"unit_scale_factor": self.unit_scale_factor})
        return variables

    def equation_lambda_eval(self, input_type="quantity"):
        """Evaluates the equation lambda function using either the quantity
        (value * units), the value, or the units.
        The values and units are evaluated separately and then combined
        because some expressions don't work well
        with pint quantity types.
        """
        # This is an attempt to make the equation lambda work with pint quantities
        # note - throws an error when it doesn't return a float
        # (happens when it returns 0 from sign function, for instance)
        self._equation_quantity = self.equation_lambda(**self.equation_variables)
        if input_type == "quantity":
            return self._equation_quantity
        elif input_type == "value":
            return self._equation_quantity.magnitude
        elif input_type == "units":
            return unit_to_quantity(self._equation_quantity.units)

    # Seems like setting this as a @property doesn't cause fenics to recompile

    @property
    def form(self):
        """-1 factor because terms are defined as if they were on the
        lhs of the equation :math:`F(u;v)=0`"""
        return (
            d.Constant(-1)
            * self.equation_lambda_eval(input_type="value")
            * self.destination_species.v
            * self.measure
        )

    @property
    def scalar_form(self):
        """
        Defines scalar form for given flux.
        If the destination species is a vector function,
        the assembled form will be a vector of size NDOF.
        """
        return (
            d.Constant(-1)
            * self.equation_lambda_eval(input_type="value")
            * self.destination_species.vscalar
            * self.measure
        )

    @property
    def assembled_flux(self):
        """Attempt to convert flux units to molecules_per_second for printing."""
        try:
            self._assembled_flux = -1 * (
                d.assemble(self.scalar_form).sum() * self.equation_units * self.measure_units
            ).to(unit.molecule / unit.s)
        except Exception:
            self._assembled_flux = -1 * (
                d.assemble(self.scalar_form).sum() * self.equation_units * self.measure_units
            )
        return self._assembled_flux

    def _post_init_get_is_linear_comp(self):
        """
        If the flux is linear in terms of a compartment vector (e.g.
        :code:`dj/du['pm']`),
        then sets :code:`self.is_lienar_wrt_comp[comp_name]` to True
        """
        umap = {}

        for species_name, species in self.species.items():
            comp_name = species.compartment_name
            umap.update({species_name: "u" + comp_name})

        uset = set(umap.values())
        new_eqn = self.equation.subs(umap)

        for comp_name in self.compartments.keys():
            d_new_eqn = sym.diff(new_eqn, "u" + comp_name, 1)
            d_new_eqn_species = {str(x) for x in d_new_eqn.free_symbols}
            self.is_linear_wrt_comp[comp_name] = uset.isdisjoint(d_new_eqn_species)


class FormContainer(ObjectContainer):
    def __init__(self):
        super().__init__(Form)

        self.properties_to_print = ["form", "form_type", "_compartment_name"]

    def print(self, tablefmt="fancy_grid", properties_to_print=None):
        for f in self:
            f.integrals
        super().print(tablefmt, self.properties_to_print)


@dataclass
class Form(ObjectInstance):
    """
    form_type:
    'mass': transient/mass form (holds time derivative)
    'diffusion'
    'domain_reaction'
    'boundary_reaction'

    Differentiating using ufl doesn't seem to get it right when
    using vector functions. Luckily we have all fluxes as sympy objects
    and mass/diffusive forms are always linear w.r.t components.
    """

    name: str
    form_: ufl.Form
    species: Species
    form_type: str
    units: pint.Unit
    is_lhs: bool
    linear_wrt_comp: dict = dataclasses.field(default_factory=dict)
    form_scaling: float = 1.0

    def set_scaling(self, form_scaling=1.0, print_scaling=True):
        self.form_scaling = form_scaling
        self.form_scaling_dolfin_constant.assign(self.form_scaling)
        if print_scaling:
            logger.info(
                f"Form scaling for form {self.name} set to {self.form_scaling}",
                extra=dict(format_type="log"),
            )

    @property
    def form(self):
        return self.form_scaling_dolfin_constant * self.form_

    @property
    def lhs(self):
        if self.is_lhs:
            return self.form
        else:
            return d.Constant(-1) * self.form

    @property
    def rhs(self):
        if self.is_lhs:
            return d.Constant(-1) * self.form
        else:
            return self.form

    def __post_init__(self):
        self.compartment = self.species.compartment
        self._compartment_name = self.compartment.name

        self.form_scaling_dolfin_constant = d.Constant(self.form_scaling)

        self._convert_pint_quantity_to_unit()
        self._check_input_type_validity()
        self._convert_pint_unit_to_quantity()

    @property
    def integrals(self):
        self._integrals = self.form.integrals()
        return self._integrals


def empty_sbmodel():
    """Initialize empty containers (pc, sc, cc, and rc)"""
    pc = ParameterContainer()
    sc = SpeciesContainer()
    cc = CompartmentContainer()
    rc = ReactionContainer()
    return pc, sc, cc, rc


def sbmodel_from_locals(local_values):
    """Assemble containers from local variables"""
    # FIXME: Add typing
    # Initialize containers
    pc, sc, cc, rc = empty_sbmodel()
    parameters = [x for x in local_values if isinstance(x, Parameter)]
    species = [x for x in local_values if isinstance(x, Species)]
    compartments = [x for x in local_values if isinstance(x, Compartment)]
    reactions = [x for x in local_values if isinstance(x, Reaction)]
    # we just reverse the list so that the order is the same as how they were defined
    parameters.reverse()
    species.reverse()
    compartments.reverse()
    reactions.reverse()
    pc.add(parameters)
    sc.add(species)
    cc.add(compartments)
    rc.add(reactions)
    return pc, sc, cc, rc



from . import config
from . import common, mesh, model, model_assembly, solvers, units, mesh_tools

__all__ = [
    "config",
    "model_assembly",
    "solvers",
    "model",
    "common",
    "mesh",
    "units",
    "mesh_tools",
]



import math
import numpy as np
import pytest
import sympy as sym

import smart


def test_Parameter_initialization(parameter_kwargs_k3f):
    """Test that we can initialize k3f SpeciesContainer"""
    k3f = smart.model_assembly.Parameter(**parameter_kwargs_k3f)

    assert k3f.name == parameter_kwargs_k3f["name"]
    assert math.isclose(k3f.value, parameter_kwargs_k3f["value"])
    assert np.allclose(k3f.value_vector, [0.0, parameter_kwargs_k3f["value"]])

    assert k3f.unit == parameter_kwargs_k3f["unit"]
    assert k3f.quantity == parameter_kwargs_k3f["value"] * parameter_kwargs_k3f["unit"]

    assert k3f.type == "constant"
    assert k3f.group == parameter_kwargs_k3f["group"]
    assert k3f.use_preintegration is parameter_kwargs_k3f["use_preintegration"]
    assert k3f.is_space_dependent is False
    assert k3f.is_time_dependent is False
    assert k3f.notes == parameter_kwargs_k3f["notes"]


@pytest.mark.parametrize("use_preintegration", [True, False])
def test_Parameter_from_expression(use_preintegration):
    Vmax, t0, m = 500, 0.1, 200
    t = sym.symbols("t")
    pulseI = Vmax * sym.atan(m * (t - t0))
    pulse = sym.diff(pulseI, t)
    value = Vmax * m / (1 + (m * (0.0 - t0)) ** 2)

    flux_unit = smart.units.unit.molecule / (smart.units.unit.um**2 * smart.units.unit.sec)
    j1pulse = smart.model_assembly.Parameter.from_expression(
        "j1pulse",
        pulse,
        flux_unit,
        use_preintegration=use_preintegration,
        preint_sym_expr=pulseI,
    )

    assert math.isclose(j1pulse.value, value)
    assert np.allclose(j1pulse.value_vector, [0.0, value])

    assert j1pulse.unit == smart.units.unit_to_quantity(flux_unit)
    assert np.isclose(j1pulse.quantity, value * flux_unit)

    assert j1pulse.type == "expression"
    assert j1pulse.use_preintegration is use_preintegration
    assert j1pulse.is_space_dependent is False
    assert j1pulse.is_time_dependent is True


def test_ParameterContainer(parameter_kwargs_k3f):
    """Test that we can initialize k3f SpeciesContainer"""
    k3f = smart.model_assembly.Parameter(**parameter_kwargs_k3f)
    k3r = smart.model_assembly.Parameter("k3r", 100, 1 / smart.units.unit.sec)
    pc = smart.model_assembly.ParameterContainer()
    assert pc.size == 0
    pc.add([k3f])
    assert pc.size == 1
    pc["k3f"] == k3f
    pc.add([k3f])
    # Adding same Parameter should not do anything
    assert pc.size == 1
    pc.add([k3r])
    assert pc.size == 2
    assert set(pc.keys) == {"k3f", "k3r"}



import math

import pytest

import smart


def test_Species_initialization(species_kwargs_A):
    """Test that we can initialize a SpeciesContainer"""
    A = smart.model_assembly.Species(**species_kwargs_A)
    assert A.name == species_kwargs_A["name"]
    assert A.latex_name == species_kwargs_A["name"]
    assert str(A.sym) == species_kwargs_A["name"]

    assert math.isclose(A.initial_condition, species_kwargs_A["initial_condition"])
    assert (
        A.initial_condition_quantity
        == species_kwargs_A["initial_condition"] * species_kwargs_A["concentration_units"]
    )
    assert A.concentration_units == smart.units.unit_to_quantity(
        species_kwargs_A["concentration_units"]
    )
    assert math.isclose(A.D, species_kwargs_A["D"])
    assert A.diffusion_units == smart.units.unit_to_quantity(species_kwargs_A["diffusion_units"])
    assert A.D_quantity == species_kwargs_A["D"] * species_kwargs_A["diffusion_units"]

    assert A.compartment_name == species_kwargs_A["compartment_name"]
    assert A.group == species_kwargs_A["group"]
    assert A.dof_map is None
    assert A.ut is None
    assert A.v is None
    assert A.u == {}
    assert A.is_an_added_species is False
    assert A.is_in_a_reaction is False


def test_SpeciesContainer(species_kwargs_A, species_kwargs_AER):
    """Test that we can initialize a SpeciesContainer"""
    A = smart.model_assembly.Species(**species_kwargs_A)
    AER = smart.model_assembly.Species(**species_kwargs_AER)
    sc = smart.model_assembly.SpeciesContainer()
    assert sc.size == 0
    sc.add([A])
    assert sc.size == 1
    sc["A"] == A
    sc.add([A])
    # Adding same species should not do anything
    assert sc.size == 1
    sc.add([AER])
    assert sc.size == 2
    assert set(sc.keys) == {"A", "AER"}


def test_add_non_Species_to_SpeciesContainer_raises_InvalidObjectException(
    compartment_kwargs_Cyto,
):
    """Test that if we try to add something different than a species
    to a SpeciesContainer then an exception will be raised"""
    sc = smart.model_assembly.SpeciesContainer()
    Cyto = smart.model_assembly.Compartment(**compartment_kwargs_Cyto)
    with pytest.raises(smart.model_assembly.InvalidObjectException):
        sc.add([Cyto])



import warnings
import functools

__all__ = ["deprecated"]


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter("always", DeprecationWarning)  # turn off filter
        warnings.warn(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)  # reset filter
        raise RuntimeError(f"Call to deprecated function {func.__name__}")
        # return func(*args, **kwargs)

    return new_func



"""
Utility functions for visualizing SMART meshes using pyvista.
"""
import dolfin
import pyvista
from ufl import vertex
from typing import Tuple
import numpy.typing as npt
import numpy as np
import pathlib
from typing import Optional

# pyvista.global_theme.trame.server_proxy_enabled = True

__all__ = ["create_vtk_structures", "plot"]

_vtk_perm = {
    dolfin.interval: {i: np.arange(i) for i in range(2, 10)},
    vertex: {1: 1},
    dolfin.triangle: {
        1: [0, 1, 2],
        2: [0, 1, 2, 5, 3, 4],
        3: [0, 1, 2, 7, 8, 3, 4, 6, 5, 9],
        4: [0, 1, 2, 9, 10, 11, 3, 4, 5, 8, 7, 6, 12, 13, 14],
    },
    dolfin.tetrahedron: {
        1: [0, 1, 2, 3],
        2: [0, 1, 2, 3, 9, 6, 8, 7, 5, 4],
        3: [0, 1, 2, 3, 14, 15, 8, 9, 13, 12, 10, 11, 6, 7, 4, 5, 18, 16, 17, 19],
    },
}


def create_vtk_structures(
    Vh: dolfin.FunctionSpace,
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.int32], npt.NDArray[np.float64]]:
    """Given a (discontinuous) Lagrange space, create pyvista compatible structures based on the
    dof coordinates

    Args:
        Vh: the function space
    Returns:
        Tuple: Mesh topology, cell types and mesh geometry arrays
    """
    family = Vh.ufl_element().family()
    mesh = Vh.mesh()

    num_cells = mesh.num_cells()
    if num_cells == 0:
        return (
            np.zeros(0, dtype=np.int32),
            np.zeros(0, dtype=np.int32),
            np.zeros((0, 3), dtype=np.float64),
        )
    d_cell = mesh.ufl_cell()
    # need this extra statement for case of triangles in 3d
    if d_cell._cellname == "triangle":
        d_cell = dolfin.triangle

    if d_cell == vertex:
        cell_type = 1
    elif d_cell == dolfin.interval:
        cell_type = 68
    elif d_cell == dolfin.triangle:
        cell_type = 69
    elif d_cell == dolfin.tetrahedron:
        cell_type = 71
    else:
        raise RuntimeError(f"Unsupported {d_cell=}")
    cell_types = np.full(num_cells, cell_type, dtype=np.int32)
    bs = 1 if Vh.num_sub_spaces() == 0 else Vh.num_sub_spaces()
    Vh = Vh if bs == 1 else Vh.sub(0).collapse()
    num_dofs_per_cell = len(Vh.dofmap().cell_dofs(0))
    topology = np.zeros((num_cells, num_dofs_per_cell + 1), dtype=np.int32)
    topology[:, 0] = num_dofs_per_cell
    dofmap = Vh.dofmap()
    num_vertices_local = dofmap.index_map().size(dofmap.index_map().MapSize.ALL)

    x = np.zeros((num_vertices_local, 3), dtype=np.float64)
    el = Vh.element()
    for i in range(num_cells):
        topology[i, 1:] = dofmap.cell_dofs(i)
        cell = dolfin.Cell(mesh, i)
        cell_dof_coords = el.tabulate_dof_coordinates(cell)
        for j, d in enumerate(dofmap.cell_dofs(i)):
            x[d, : cell_dof_coords.shape[1]] = cell_dof_coords[j]

    degree = Vh.ufl_element().degree()
    try:
        perm = _vtk_perm[d_cell][degree]
    except KeyError:
        raise RuntimeError(f"Unsupported plotting of space {family} of {degree=} on {d_cell}")
    topology[:, 1:] = topology[:, 1:][:, perm]
    return topology, cell_types, x


def plot(
    uh: dolfin.Function,
    filename: Optional[pathlib.Path] = None,
    show_edges: bool = True,
    glyph_factor: float = 1,
    off_screen: bool = True,
    view_xy: bool = False,
):
    """
    Plot a (discontinuous) Lagrange function with Pyvista

    Args:
        uh: The function
        filename: If set, writes the plot to file instead of displaying it interactively
        show_edges: Show mesh edges if ``True``
        glyph_factor: Scaling of glyphs if input function is a function from a
            ``dolfin.VectorFunctionSpace``.
        off_screen: If ``True`` generate plots with virtual frame buffer using ``xvfb``.
        view_xy: If ``True``, view xy plane
    """
    Vh = uh.function_space()

    dofmap = Vh.dofmap()
    num_vertices_local = dofmap.index_map().size(dofmap.index_map().MapSize.ALL)
    bs = 1 if Vh.num_sub_spaces() == 0 else Vh.num_sub_spaces()

    u_vec = uh.vector().get_local(np.arange(num_vertices_local))
    if bs > 1:
        u_out = np.zeros((len(u_vec) // bs, 3), dtype=np.float64)
        u_out[:, :bs] = u_vec.reshape(len(u_vec) // bs, bs)
    else:
        u_out = u_vec
    topology, cell_types, x = create_vtk_structures(Vh)

    grid = pyvista.UnstructuredGrid(topology, cell_types, x)
    grid[uh.name()] = u_out

    if bs > 1:
        grid.set_active_scalars(None)

    pyvista.OFF_SCREEN = off_screen
    pyvista.start_xvfb()

    plotter = pyvista.Plotter()

    # Workaround for pyvista issue for higherorder grids
    degree = Vh.ufl_element().degree()
    if degree > 1 and show_edges:
        first_order_el = Vh.ufl_element().reconstruct(degree=1)
        P1 = dolfin.FunctionSpace(Vh.mesh(), first_order_el)
        t1, c1, x1 = create_vtk_structures(P1)
        p1_grid = pyvista.UnstructuredGrid(t1, c1, x1)
        if bs > 1:
            plotter.add_mesh(p1_grid, show_edges=show_edges)
        else:
            plotter.add_mesh(grid)
            plotter.add_mesh(p1_grid, style="wireframe")
    else:
        if uh.geometric_dimension() == 2 or view_xy:
            plotter.add_mesh(grid, show_edges=show_edges)
        else:
            crinkled = grid.clip(normal=(1, 0, 0), crinkle=True)
            plotter.add_mesh(crinkled, show_edges=show_edges)

    if uh.geometric_dimension() == 2 or view_xy:
        plotter.view_xy()

    if bs > 1:
        glyphs = grid.glyph(orient=uh.name(), factor=glyph_factor)
        plotter.add_mesh(glyphs)
    print(filename)
    if filename is None:
        plotter.show()
    else:
        plotter.screenshot(filename)


def plot_dolfin_mesh(
    msh: dolfin.mesh,
    mf_cell: dolfin.MeshFunction,
    mf_facet: dolfin.MeshFunction = None,
    outer_marker: int = 10,
    filename: Optional[pathlib.Path] = None,
    show_edges: bool = True,
    off_screen: bool = True,
    view_xy: bool = False,
):
    """
    Construct P1 function space on current mesh,
    convert function space to vtk structures,
    and then plot mesh markers in mf (a dolfin MeshFunction)

    Args:
        msh: dolfin mesh object
        mf_cell: marker values for cells in domain, given as a dolfin MeshFunction
        mf_facet (optional): marker values for facets at domain boundary,
            given as a dolfin MeshFunction defined over all facets in the mesh
        outer_marker (optional): value marking the boundary facets in mf_facet.
            Number of nodes with this marker
            should match the number of nodes in dolfin.BoundaryMesh(msh, "exterior")
        filename: If set, writes the plot to file instead of displaying it interactively
        show_edges: Show mesh edges if ``True``
        off_screen: If ``True`` generate plots with virtual frame buffer using ``xvfb``.
        view_xy: If ``True``, view xy plane
    """
    Vh = dolfin.FunctionSpace(msh, "P", 1)
    u_out = mf_cell.array()
    topology, cell_types, x = create_vtk_structures(Vh)
    grid = pyvista.UnstructuredGrid(topology, cell_types, x)
    grid["mf"] = u_out

    facets_loaded = False
    if mf_facet is not None:
        msh_facet = dolfin.BoundaryMesh(msh, "exterior")
        Vh_facet = dolfin.FunctionSpace(msh_facet, "P", 1)
        u_facet = mf_facet.array()[
            mf_facet.array() == outer_marker
        ]  # outer_marker labels outer surface
        topology, cell_types, x = create_vtk_structures(Vh_facet)
        grid_facet = pyvista.UnstructuredGrid(topology, cell_types, x)
        try:
            grid_facet["mf"] = u_facet
            facets_loaded = True
        except ValueError:
            facets_loaded = False

    pyvista.OFF_SCREEN = off_screen
    pyvista.start_xvfb()

    plotter = pyvista.Plotter()

    # Workaround for pyvista issue for higherorder grids
    degree = Vh.ufl_element().degree()
    if degree > 1 and show_edges:
        first_order_el = Vh.ufl_element().reconstruct(degree=1)
        P1 = dolfin.FunctionSpace(Vh.mesh(), first_order_el)
        t1, c1, x1 = create_vtk_structures(P1)
        p1_grid = pyvista.UnstructuredGrid(t1, c1, x1)
        plotter.add_mesh(grid)
        plotter.add_mesh(p1_grid, style="wireframe")
    else:
        if msh.topology().dim() == 3:
            crinkled = grid.clip(normal=(1, 0, 0), crinkle=True)
            plotter.add_mesh(crinkled, show_edges=show_edges)
            if facets_loaded:
                crinkled_facet = grid_facet.clip(normal=(1, 0, 0), crinkle=True)
                plotter.add_mesh(crinkled_facet, show_edges=show_edges)
        elif msh.topology().dim() == 2:
            plotter.add_mesh(grid, show_edges=show_edges)
            if facets_loaded:
                plotter.add_mesh(grid_facet, show_edges=show_edges)

    if msh.geometric_dimension() == 2 or view_xy:
        plotter.view_xy()

    print(filename)
    if filename is None:
        plotter.show()
    else:
        plotter.screenshot(filename)



import smart


def test_Compartment_initialization(compartment_kwargs_Cyto):
    """Test that we can initialize a Compartment"""
    Cyto = smart.model_assembly.Compartment(**compartment_kwargs_Cyto)

    assert Cyto.V is None
    assert Cyto.cell_marker == compartment_kwargs_Cyto["cell_marker"]
    assert Cyto.dimensionality == compartment_kwargs_Cyto["dimensionality"]
    assert Cyto.name == compartment_kwargs_Cyto["name"]
    assert Cyto.num_dofs == 0
    assert Cyto.num_dofs_local == 0
    assert Cyto.species == {}
    assert Cyto.u == {}
    assert Cyto.v is None


def test_CompartmentContainer(compartment_kwargs_Cyto, compartment_kwargs_PM):
    """Test that we can initialize a CompartmentContainer"""
    Cyto = smart.model_assembly.Compartment(**compartment_kwargs_Cyto)
    PM = smart.model_assembly.Compartment(**compartment_kwargs_PM)
    cc = smart.model_assembly.CompartmentContainer()
    assert cc.size == 0
    cc.add([Cyto])
    assert cc.size == 1
    cc["Cyto"] == Cyto
    cc.add([Cyto])
    # Adding same species should not do anything
    assert cc.size == 1
    cc.add([PM])
    assert cc.size == 2
    assert set(cc.keys) == {"Cyto", "PM"}



"""
General functions: array manipulation, data i/o, etc
"""

import time
import logging
from datetime import datetime
from typing import Any, Callable, Dict, List, Union

import dolfin as d

import ufl
from pytz import timezone
from termcolor import colored

from .deprecation import deprecated
from .config import global_settings as gset, FancyFormatter
from .units import unit

__all__ = [
    "smart_expressions",
    "sub",
    "Stopwatch",
]


comm = d.MPI.comm_world
rank = comm.rank
size = comm.size
root = 0

logger = logging.getLogger(__name__)


def smart_expressions(
    dolfin_expressions: Dict[str, Callable[[Any], Any]]
) -> Dict[str, Callable[[Any], Union[ufl.core.expr.Expr, float]]]:
    """
    Map strings to DOLFIN/UFL functions, that takes in
    `smart`-Expressions, i.e. functions with a unit.

    Args:
        dolfin_expressions: Dictionary of strings mapping
          to smart expressions

    Example:

        .. highlight:: python
        .. code-block:: python

            input = {"sin": ufl.sin}
            output = smart_expressions(input)

        Output is then a dictionary that maps "sin" to
        a function :code:`sin(x)` that takes in a
        function with a unit and returns
        :code:`ufl.sin(x.to(unit.dimensionless).magnitude)`
    """
    # must use a function here or else all functions are defined as the final "v"
    # see e.g. https://stackoverflow.com/questions/36805071
    def dict_entry(v):
        return lambda x: v(x.to(unit.dimensionless).magnitude)

    return {k: dict_entry(v) for k, v in dolfin_expressions.items()}


def sub(
    func: Union[
        List[Union[d.Function, d.FunctionSpace]],
        d.Function,
        d.MixedFunctionSpace,
        d.FunctionSpace,
        d.function.argument.Argument,
    ],
    idx: int,
    collapse_function_space: bool = True,
):
    """
    A utility function for getting the sub component of a

    - Functionspace
    - Mixed function space
    - Function
    - Test/Trial Function

    Args:
        func: The input
        idx: The component to extract
        collapse_function_space: If the input `func`
        is a `dolfin.FunctionSpace`, collapse it if `True`.

    Returns:
        The relevant component of the input
    """

    if isinstance(func, (list, tuple)):
        return func[idx]

    elif isinstance(func, d.Function):
        # MixedFunctionSpace
        if func._functions:
            assert func.sub(idx) is func._functions[idx]
            return func.sub(idx)
        else:
            # scalar FunctionSpace
            if func.num_sub_spaces() <= 1 and idx == 0:
                return func
            else:
                # Use d.split() to get subfunctions from a
                # VectorFunctionSpace that can be used in forms
                return func.sub(idx)

    elif isinstance(func, d.MixedFunctionSpace):
        return func.sub(idx)

    elif isinstance(func, d.FunctionSpace):
        if func.num_sub_spaces() <= 1 and idx == 0:
            return func
        else:
            if collapse_function_space:
                return func.sub(idx).collapse()
            else:
                return func.sub(idx)

    elif isinstance(func, d.function.argument.Argument):
        if func.function_space().num_sub_spaces() <= 1 and idx == 0:
            return func
        else:
            return func[idx]

    else:
        raise ValueError(f"Unknown input type of {func=}")


# Write a stopwatch class to measure time elapsed
# with a start, stop, and pause methods
# Keep track of timings in a list of lists called
# self.timings, each time the timer is paused,
# the time elapsed since the last pause is added to
# the sublist. Using stop resets the timer to zero
# and beings a new list of timings.


class Stopwatch:
    "Basic stopwatch class with inner/outer timings (pause and stop)"

    def __init__(self, name=None, time_unit="s", print_buffer=0, filename=None, start=False):
        self.name = name
        self.time_unit = time_unit
        self.stop_timings = []  # length = number of stops
        self.pause_timings = []  # length = number of stops (list of lists)
        self._pause_timings = []  # length = number of pauses (reset on stop)
        self._times = []
        self.is_paused = True
        self.print_buffer = print_buffer
        self._print_name = f"{str(self.name): <{self.print_buffer}}"
        # self.start()
        self.filename = filename
        if filename is not None:
            self.file_logger = logging.getLogger("smart_stop_watch")
            handler = logging.FileHandler(filename=filename)
            handler.setFormatter(FancyFormatter())
            self.file_logger.addHandler(handler)
            self.file_logger.setLevel(logging.DEBUG)
        else:
            # Just use the regular logger
            self.file_logger = logger
        if start:
            self.start()

    def start(self):
        self._times.append(time.time())
        self.is_paused = False

    def pause(self):
        if self.is_paused:
            return
        else:
            self._times.append(time.time())
            self._pause_timings.append(self._times[-1] - self._times[-2])
            self.is_paused = True
            self.file_logger.debug(
                f"{self.name} (iter {len(self._pause_timings)}) finished "
                f"in {self.time_str(self._pause_timings[-1])} {self.time_unit}",
                extra=dict(format_type="logred"),
            )

    def stop(self, print_result=True):
        self._times.append(time.time())
        if self.is_paused:
            final_time = 0
        else:
            final_time = self._times[-1] - self._times[-2]
            self.is_paused = True
        total_time = sum(self._pause_timings) + final_time
        self.stop_timings.append(total_time)
        if print_result:
            self.file_logger.debug(
                f"{self._print_name} finished in {self.time_str(total_time)} {self.time_unit}",
                extra=dict(format_type="logred"),
            )

        self.pause_timings.append(self._pause_timings)
        self._pause_timings = []
        self._times = []

    def set_timing(self, timing):
        self.stop_timings.append(timing)
        self.file_logger.debug(
            f"{self._print_name} finished in {self.time_str(timing)} {self.time_unit}",
            extra=dict(format_type="logred"),
        )

    def print_last_stop(self):
        self.file_logger.debug(
            f"{self._print_name} finished in "
            f"{self.time_str(self.stop_timings[-1])} {self.time_unit}",
            extra=dict(format_type="logred"),
        )

    def time_str(self, t):
        return str({"us": 1e6, "ms": 1e3, "s": 1, "min": 1 / 60}[self.time_unit] * t)[0:8]


@deprecated
def _fancy_print(
    title_text,
    buffer_color=None,
    text_color=None,
    filler_char=None,
    num_banners=None,
    new_lines=None,
    left_justify=None,
    format_type="default",
    include_timestamp=True,
    filename=None,
):
    "Formatted text to stand out."

    # Initialize with the default options
    buffer_color_ = "cyan"
    text_color_ = "green"
    filler_char_ = "="
    num_banners_ = 0
    new_lines_ = [0, 0]
    left_justify_ = False
    # Override with format_type options
    if format_type == "default":
        pass
    elif format_type == "title":
        text_color_ = "magenta"
        num_banners_ = 1
        new_lines_ = [1, 0]
    elif format_type == "subtitle":
        text_color_ = "green"
        filler_char_ = "."
        left_justify_ = True
    elif format_type == "data":
        buffer_color_ = "white"
        text_color_ = "white"
        filler_char_ = ""
        left_justify_ = True
    elif format_type == "data_important":
        buffer_color_ = "white"
        text_color_ = "red"
        filler_char_ = ""
        left_justify_ = True
    elif format_type == "log":
        buffer_color_ = "white"
        text_color_ = "green"
        filler_char_ = ""
        left_justify_ = True
    elif format_type == "logred":
        buffer_color_ = "white"
        text_color_ = "red"
        filler_char_ = ""
        left_justify_ = True
    elif format_type == "log_important":
        buffer_color_ = "white"
        text_color_ = "magenta"
        filler_char_ = "."
    elif format_type == "log_urgent":
        buffer_color_ = "white"
        text_color_ = "red"
        filler_char_ = "."
    elif format_type == "warning":
        buffer_color_ = "magenta"
        text_color_ = "red"
        filler_char_ = "!"
        num_banners_ = 2
        new_lines_ = [1, 1]
    elif format_type == "timestep":
        text_color_ = "red"
        num_banners_ = 2
        filler_char_ = "."
        new_lines_ = [1, 1]
    elif format_type == "solverstep":
        text_color_ = "red"
        num_banners_ = 1
        filler_char_ = "."
        new_lines_ = [1, 1]
    elif format_type == "assembly":
        text_color_ = "magenta"
        num_banners_ = 0
        filler_char_ = "."
        new_lines_ = [1, 0]
    elif format_type == "assembly_sub":
        text_color_ = "magenta"
        num_banners_ = 0
        filler_char_ = ""
        new_lines_ = [0, 0]
        left_justify_ = True
    elif format_type is not None:
        raise ValueError("Unknown formatting_type.")

    # Override again with user options
    if buffer_color is None:
        buffer_color = buffer_color_
    if text_color is None:
        text_color = text_color_
    if filler_char is None:
        filler_char = filler_char_
    if num_banners is None:
        num_banners = num_banners_
    if new_lines is None:
        new_lines = new_lines_
    if left_justify is None:
        left_justify = left_justify_

    # include MPI rank in message
    if size > 1:
        title_text = f"CPU {rank}: {title_text}"
    if include_timestamp:
        timestamp = datetime.now(timezone("US/Pacific")).strftime("[%Y-%m-%d time=%H:%M:%S]")
        title_text = f"{timestamp} {title_text}"

    # calculate optimal buffer size
    min_buffer_size = 5
    terminal_width = 120
    buffer_size = max(
        [min_buffer_size, int((terminal_width - 1 - len(title_text)) / 2 - 1)]
    )  # terminal width == 80
    title_str_len = (buffer_size + 1) * 2 + len(title_text)
    parity = 1 if title_str_len == 78 else 0

    # color/stylize buffer, text, and banner
    def buffer(buffer_size):
        return colored(filler_char * buffer_size, buffer_color)

    if left_justify:
        title_str = f"{colored(title_text, text_color)} {buffer(buffer_size*2+1+parity)}"
    else:
        title_str = (
            f"{buffer(buffer_size)} {colored(title_text, text_color)} "
            f"{buffer(buffer_size+parity)}"
        )
    banner = colored(filler_char * (title_str_len + parity), buffer_color)

    def print_out(text, filename=None):
        "print to file and terminal"
        if filename is not None:
            with open(filename, "a") as f:
                f.write(text + "\n")
        elif gset["log_filename"] is not None:
            with open(gset["log_filename"], "a") as f:
                f.write(text + "\n")
        print(text)

    # initial spacing
    if new_lines[0] > 0:
        print_out("\n" * (new_lines[0] - 1), filename)
    # print first banner
    for _ in range(num_banners):
        print_out(f"{banner}", filename)
    # print main text
    print_out(title_str, filename)
    # print second banner
    for _ in range(num_banners):
        print_out(f"{banner}", filename)
    # end spacing
    if new_lines[1] > 0:
        print_out("\n" * (new_lines[1] - 1), filename)



"""
Configuration settings for simulation: logger formatting, solver configuration, other flags
"""
import logging
from logging import config as logging_config
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, Optional, Tuple, NamedTuple


from termcolor import colored
import dolfin as d
import ufl


__all__ = [
    "global_settings",
    "Config",
    "SolverConfig",
    "BaseConfig",
    "FlagsConfig",
]


comm = d.MPI.comm_world
rank = comm.rank
size = comm.size
root = 0
fancy_format = (
    "%(asctime)s %(rank)s%(name)s - %(levelname)s - "
    "%(newline)s%(message)s%(newline)s (%(filename)s:%(lineno)d)"
)
base_format = "%(asctime)s %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
root_format = "ROOT -" + base_format


class FormatType(str, Enum):
    default = "default"
    title = "title"  # type: ignore
    subtitle = "subtitle"
    data = "data"
    data_important = "data_important"
    log = "log"
    logred = "logred"
    log_important = "log_important"
    log_urgent = "log_urgent"
    warning = "warning"
    timestep = "timestep"
    solverstep = "solverstep"
    assembly = "assembly"
    assembly_sub = "assembly_sub"
    table = "table"


class FormatOption(NamedTuple):
    buffer_color: str = "cyan"
    text_color: str = "white"
    filler_char: str = ""
    num_banners: int = 0
    new_lines: Tuple[int, int] = (0, 0)
    left_justify: bool = False
    is_table: bool = False


def format_type_to_options(format_type: FormatType) -> FormatOption:
    """Take a format type and return the corresponding options

    Parameters
    ----------
    format_type : FormatType
        The format type

    Returns
    -------
    FormatOption
        The options

    Raises
    ------
    ValueError
        If provided format type does not exist.
    """
    if FormatType[format_type] == FormatType.default:
        return FormatOption()
    if FormatType[format_type] == FormatType.table:
        return FormatOption(is_table=True)
    elif FormatType[format_type] == FormatType.title:
        return FormatOption(
            text_color="magenta",
            num_banners=1,
            new_lines=(1, 0),
        )
    elif FormatType[format_type] == FormatType.subtitle:
        return FormatOption(
            text_color="green",
            filler_char="",
            left_justify=True,
        )
    elif FormatType[format_type] == FormatType.data:
        return FormatOption(
            buffer_color="white",
            text_color="white",
            filler_char="",
            left_justify=True,
        )
    elif FormatType[format_type] == FormatType.data_important:
        return FormatOption(
            buffer_color="white",
            text_color="red",
            filler_char="",
            left_justify=True,
        )
    elif FormatType[format_type] == FormatType.log:
        return FormatOption(
            buffer_color="white",
            text_color="green",
            filler_char="",
            left_justify=True,
        )
    elif FormatType[format_type] == FormatType.logred:
        return FormatOption(
            buffer_color="white",
            text_color="red",
            filler_char="",
            left_justify=True,
        )
    elif FormatType[format_type] == FormatType.log_important:
        return FormatOption(
            buffer_color="white",
            text_color="magenta",
            filler_char="",
        )
    elif FormatType[format_type] == FormatType.log_urgent:
        return FormatOption(
            buffer_color="white",
            text_color="red",
            filler_char="",
        )
    elif FormatType[format_type] == FormatType.warning:
        return FormatOption(
            buffer_color="magenta",
            text_color="red",
            filler_char="",
            num_banners=2,
            new_lines=(1, 1),
        )
    elif FormatType[format_type] == FormatType.timestep:
        return FormatOption(
            text_color="red",
            num_banners=2,
            filler_char="",
            new_lines=(1, 1),
        )
    elif FormatType[format_type] == FormatType.solverstep:
        return FormatOption(
            text_color="red",
            num_banners=1,
            filler_char="",
            new_lines=(1, 1),
        )
    elif FormatType[format_type] == FormatType.assembly:
        return FormatOption(
            text_color="magenta",
            num_banners=0,
            filler_char="",
            new_lines=(1, 0),
        )
    elif FormatType[format_type] == FormatType.assembly_sub:
        return FormatOption(
            text_color="magenta",
            num_banners=0,
            filler_char="",
            new_lines=(0, 0),
            left_justify=True,
        )
    else:
        msg = f"Invalid format type {format_type}"
        raise ValueError(msg)


def format_message(title_text: str, format_option: FormatOption) -> Tuple[str, str]:
    """Format message according to the format options

    Parameters
    ----------
    title_text : str
        The message to be formatted
    format_option : FormatOption
        The options for formatting

    Returns
    -------
    Tuple[str, str]
        (banners, formatted text)
    """
    min_buffer_size = 5
    terminal_width = 120
    buffer_size = max(
        [min_buffer_size, int((terminal_width - 1 - len(title_text)) / 2 - 1)]
    )  # terminal width == 80
    title_str_len = (buffer_size + 1) * 2 + len(title_text)
    parity = 1 if title_str_len == 78 else 0

    # color/stylize buffer, text, and banner
    def buffer(buffer_size):
        return colored(format_option.filler_char * buffer_size, format_option.buffer_color)

    if format_option.left_justify:
        title_str = (
            f"{colored(title_text, format_option.text_color)} {buffer(buffer_size*2+1+parity)}"
        )
    else:
        title_str = (
            f"{buffer(buffer_size)} {colored(title_text, format_option.text_color)} "
            f"{buffer(buffer_size+parity)}"
        )
    banner = colored(
        format_option.filler_char * (title_str_len + parity), format_option.buffer_color
    )
    return banner, title_str


class FancyFormatter(logging.Formatter):
    "Custom Formatter for logging"

    def format(self, record: logging.LogRecord) -> str:
        # Set the rank attribute with is a parameter in the `fancy_format``
        record.rank = f"CPU {rank}: " if size > 1 else ""

        # Extract the format options. Here we expect that `format_type` is provided
        # as part of the `extra` dictionary passed to the logger
        format_option = format_type_to_options(getattr(record, "format_type", FormatType.default))
        record.newline = "\n" if format_option.is_table else ""
        # Create the formatter
        formatter = logging.Formatter(fancy_format)
        # Format the record
        out = formatter.format(record)
        # Now create banners and formatted text
        banner, formatted_out = format_message(out, format_option=format_option)

        # Finally construct the output with banners and new lines
        banners = f"\n{banner}" * format_option.num_banners + "\n"
        prefix = "\n" * format_option.new_lines[0]
        postfix = ""
        if format_option.num_banners > 0:
            prefix += banners
            postfix += banners
        postfix += "\n" * format_option.new_lines[1]
        return prefix + formatted_out + postfix


LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "root": {"format": root_format},
        "base": {"format": base_format},
        "default": {"()": FancyFormatter},
    },
    "handlers": {
        "console": {"class": "logging.StreamHandler", "formatter": "default"},
        "root_console": {"class": "logging.StreamHandler", "formatter": "root"},
    },
    "loggers": {
        "smart": {
            "handlers": ["console"],
            "level": "DEBUG",
            # Don't send it up my namespace for additional handling
            "propagate": False,
        },
        "pint": {"level": "ERROR"},
        "FFC": {"level": "WARNING"},
        "UFL": {"level": "WARNING"},
        "dolfin": {"level": "INFO"},
        "dijitso": {"level": "INFO"},
    },
    "root": {"handlers": ["root_console"], "level": "INFO"},
}


logging_config.dictConfig(LOGGING_CONFIG)


global_settings = {
    "main_dir": None,
    "log_filename": None,
    # These functions will be substituted into any expressions
    "dolfin_expressions": {
        "exp": d.exp,
        "cos": d.cos,
        "sin": d.sin,
        "tan": d.tan,
        "cosh": ufl.cosh,
        "sinh": ufl.sinh,
        "tanh": ufl.tanh,
        "acos": d.acos,
        "asin": d.asin,
        "atan": d.atan,
        "atan2": ufl.atan_2,
        "sqrt": d.sqrt,
        "ln": d.ln,
        "abs": ufl.algebra.Abs,
        "sign": ufl.sign,
        "pi": d.pi,
        "erf": d.erf,
    },
}


class BaseConfig:
    """
    Base-class for setting configuration
    """

    def update(self, values: Dict[str, Any]):
        """
        Given a dictionary of input keys and values, update the named tuple.

        :throws AttributeError: If input key does not exist
        """
        for value, item in values.items():
            self.__setattr__(value, item)

    def __setattr__(self, key: str, value: Any):
        if key not in self.__annotations__:
            raise AttributeError(f"{key} not defined in {self.__annotations__.keys()}")
        else:
            object.__setattr__(self, key, value)

    def __getitem__(self, key):
        return self.__getattribute__(key)


@dataclass
class SolverConfig(BaseConfig):
    """
    Parameters for solver.

    Args:
        final_t: End time of simulation
        use_snes: Use PETScSNES solver if true, else use DOLFINs NewtonSolver
        snes_preassemble_linear_system: If True separate linear components during assembly
        initial_dt: Initial time-stepping
        adjust_dt: A tuple (t, dt) of floats indicating when to next adjust the
            time-stepping and to what value
        dt: Number of digits for rounding `dt`
        print_assembly: Print information during assembly process
        dt_decrease_factor:
        dt_increase_factor:
        attempt_timestep_restart_on_divergence: Restart snes solver if it diverges
        reset_timestep_for_negative_solution: Reduce solver timestep is solution is negative
    """

    final_t: Optional[float] = None
    use_snes: bool = True
    snes_preassemble_linear_system: bool = False  #: .. warning:: FIXME Currently untested
    initial_dt: Optional[float] = None
    adjust_dt: Optional[Tuple[float, float]] = None
    time_precision: int = 6
    print_assembly: bool = True
    attempt_timestep_restart_on_divergence: bool = False
    reset_timestep_for_negative_solution: bool = False


@dataclass
class FlagsConfig(BaseConfig):
    """
    Various flags

    Args:
        allow_unused_components: Allow parameters not defined in any reaction to be
            defined in any model.
        print_verbose_info: Print detailed information about a model
    """

    allow_unused_components: bool = False
    print_verbose_info: bool = True


@dataclass
class Config:
    """
    Configuration settings.

    Args:
        solver: Options for the solvers
        flags: Various options
    """

    solver: SolverConfig = field(default_factory=SolverConfig)
    flags: FlagsConfig = field(default_factory=FlagsConfig)

    @property
    def reaction_database(self) -> Dict[str, str]:
        """
        Return database of known reactions
        """
        return {"prescribed": "k", "prescribed_linear": "k*u"}



"""
Functions to create meshes for demos.

:func:`create_cubes` defines a 'cube-in-a-cube' mesh using the
built in :class:`dolfin.UnitCubeMesh()`, with subdomains defined and
marked by :func:`facet_topology` and :func:`cube_condition`
:func:`create_spheres`, :func:`create_ellipsoids`,
:func:`create_cylinders`, and :func:`create_ellipses`
define meshes using gmsh, which are then converted to
dolfin meshes using :func:`gmsh_to_dolfin`
:func:`write_mesh` writes 3d meshes with cell markers ``mf3`` and
facet markers ``mf2`` to hdf5 and pvd files.
"""

from typing import Tuple
import pathlib
import numpy as np
import dolfin as d
from mpi4py import MPI

__all__ = [
    "facet_topology",
    "cube_condition",
    "create_cubes",
    "create_spheres",
    "create_ellipsoids",
    "create_cylinders",
    "create_ellipses",
    "gmsh_to_dolfin",
    "write_mesh",
]


def facet_topology(f: d.Facet, mf3: d.MeshFunction):
    """
    Given a facet and cell mesh function,
    return the topology of the face
    as either 'boundary' (outer boundary),
    'internal', or 'interface' (boundary of inner cube)
    """
    # cells adjacent face
    localCells = [mf3.array()[c.index()] for c in d.cells(f)]
    if len(localCells) == 1:
        topology = "boundary"  # boundary facet
    elif len(localCells) == 2 and localCells[0] == localCells[1]:
        topology = "internal"  # internal facet
    elif len(localCells) == 2:
        topology = "interface"  # interface facet
    else:
        raise Exception("Facet has more than two cells")
    return (topology, localCells)


def cube_condition(cell, xmin=0.3, xmax=0.7):
    """
    Returns true when inside an inner cube region defined as:
    xmin <= x <= xmax, xmin <= y <= xmax, xmin <= z <= xmax
    """
    return (
        (xmin - d.DOLFIN_EPS < cell.midpoint().x() < xmax + d.DOLFIN_EPS)
        and (xmin - d.DOLFIN_EPS < cell.midpoint().y() < xmax + d.DOLFIN_EPS)
        and (xmin - d.DOLFIN_EPS < cell.midpoint().z() < xmax + d.DOLFIN_EPS)
    )


def create_cubes(N=16, condition=cube_condition):
    """
    Creates a mesh for use in examples that contains
    two distinct cube subvolumes with a shared interface surface.
    Cell markers:
    1 - Default subvolume (volume outside the inner cube)
    2 - Subvolume specified by condition function

    Facet markers:
    12 - Interface between subvolumes
    10 - Boundary of subvolume 1
    0  - Interior facets
    """
    # Create a mesh
    mesh = d.UnitCubeMesh(N, N, N)
    # Initialize mesh functions
    mf3 = d.MeshFunction("size_t", mesh, 3, 0)
    mf2 = d.MeshFunction("size_t", mesh, 2, 0)

    # Mark all cells that satisfy condition as 3, else 1
    for c in d.cells(mesh):
        mf3[c] = 2 if condition(c) else 1

    # Mark facets
    for f in d.faces(mesh):
        topology, cellIndices = facet_topology(f, mf3)
        if topology == "interface":
            mf2[f] = 12
        elif topology == "boundary":
            mf2[f] = int(cellIndices[0] * 10)
        else:
            mf2[f] = 0
    return (mesh, mf2, mf3)


def create_spheres(
    outerRad: float = 0.5,
    innerRad: float = 0.25,
    hEdge: float = 0,
    hInnerEdge: float = 0,
    interface_marker: int = 12,
    outer_marker: int = 10,
    inner_vol_tag: int = 2,
    outer_vol_tag: int = 1,
    comm: MPI.Comm = d.MPI.comm_world,
    verbose: bool = False,
) -> Tuple[d.Mesh, d.MeshFunction, d.MeshFunction]:
    """
    Calls create_ellipsoids() to make spherical mesh
    Args:
        outerRad: radius of the outer sphere
        innerRad: radius of the inner sphere
    All other arguments are the same as described for create_ellipsoids()
    """
    dmesh, mf2, mf3 = create_ellipsoids(
        (outerRad, outerRad, outerRad),
        (innerRad, innerRad, innerRad),
        hEdge,
        hInnerEdge,
        interface_marker,
        outer_marker,
        inner_vol_tag,
        outer_vol_tag,
        comm,
        verbose,
    )
    return (dmesh, mf2, mf3)


def create_ellipsoids(
    outerRad: Tuple[float, float, float],
    innerRad: Tuple[float, float, float],
    hEdge: float = 0,
    hInnerEdge: float = 0,
    interface_marker: int = 12,
    outer_marker: int = 10,
    inner_vol_tag: int = 2,
    outer_vol_tag: int = 1,
    comm: MPI.Comm = d.MPI.comm_world,
    verbose: bool = False,
) -> Tuple[d.Mesh, d.MeshFunction, d.MeshFunction]:
    """
    Creates a mesh for use in examples that contains
    two distinct ellipsoid subvolumes with a shared interface
    surface. If the radius of the inner ellipsoid is 0, mesh a
    single ellipsoid.

    Args:
        outerRad: The radius of the outer ellipsoid
        innerRad: The radius of the inner ellipsoid
        hEdge: maximum mesh size at the outer edge
        hInnerEdge: maximum mesh size at the edge
            of the inner ellipsoid
        interface_marker: The value to mark facets on the interface with
        outer_marker: The value to mark facets on the outer ellipsoid with
        inner_vol_tag: The value to mark the inner ellipsoidal volume with
        outer_vol_tag: The value to mark the outer ellipsoidal volume with
        comm: MPI communicator to create the mesh with
        verbose: If true print gmsh output, else skip
    Returns:
        Tuple (mesh, facet_marker, cell_marker)
    """
    import gmsh

    if np.any(np.isclose(outerRad, 0)):
        ValueError("One of the outer radii is equal to zero")
    if np.isclose(hEdge, 0):
        hEdge = 0.1 * max(outerRad)
    if np.isclose(hInnerEdge, 0):
        hInnerEdge = 0.2 * max(outerRad) if np.any(np.isclose(innerRad, 0)) else 0.2 * max(innerRad)
    if innerRad[0] > outerRad[0] or innerRad[1] > outerRad[1] or innerRad[2] > outerRad[2]:
        ValueError("Inner ellipsoid does not fit inside outer ellipsoid")
    # Create the two ellipsoid mesh using gmsh
    gmsh.initialize()
    gmsh.option.setNumber("General.Terminal", int(verbose))

    gmsh.model.add("twoellipsoids")
    # first add ellipsoid 1 of radius outerRad and center (0,0,0)
    outer_ellipsoid = gmsh.model.occ.addSphere(0, 0, 0, 1.0)
    gmsh.model.occ.dilate([(3, outer_ellipsoid)], 0, 0, 0, outerRad[0], outerRad[1], outerRad[2])
    if np.any(np.isclose(innerRad, 0)):
        # Use outer_ellipsoid only
        gmsh.model.occ.synchronize()
        gmsh.model.add_physical_group(3, [outer_ellipsoid], tag=outer_vol_tag)
        facets = gmsh.model.getBoundary([(3, outer_ellipsoid)])
        assert len(facets) == 1
        gmsh.model.add_physical_group(2, [facets[0][1]], tag=outer_marker)
    else:
        # Add inner_ellipsoid (radius innerRad, center (0,0,0))
        inner_ellipsoid = gmsh.model.occ.addSphere(0, 0, 0, 1.0)
        gmsh.model.occ.dilate(
            [(3, inner_ellipsoid)], 0, 0, 0, innerRad[0], innerRad[1], innerRad[2]
        )
        # Create interface between ellipsoids
        two_ellipsoids, (outer_ellipsoid_map, inner_ellipsoid_map) = gmsh.model.occ.fragment(
            [(3, outer_ellipsoid)], [(3, inner_ellipsoid)]
        )
        gmsh.model.occ.synchronize()

        # Get the outer boundary
        outer_shell = gmsh.model.getBoundary(two_ellipsoids, oriented=False)
        assert len(outer_shell) == 1
        # Get the inner boundary
        inner_shell = gmsh.model.getBoundary(inner_ellipsoid_map, oriented=False)
        assert len(inner_shell) == 1
        # Add physical markers for facets
        gmsh.model.add_physical_group(outer_shell[0][0], [outer_shell[0][1]], tag=outer_marker)
        gmsh.model.add_physical_group(inner_shell[0][0], [inner_shell[0][1]], tag=interface_marker)

        # Physical markers for
        all_volumes = [tag[1] for tag in outer_ellipsoid_map]
        inner_volume = [tag[1] for tag in inner_ellipsoid_map]
        outer_volume = []
        for vol in all_volumes:
            if vol not in inner_volume:
                outer_volume.append(vol)
        gmsh.model.add_physical_group(3, outer_volume, tag=outer_vol_tag)
        gmsh.model.add_physical_group(3, inner_volume, tag=inner_vol_tag)

    def meshSizeCallback(dim, tag, x, y, z, lc):
        # mesh length is hEdge at the PM (defaults to 0.1*outerRad,
        # or set when calling function) and hInnerEdge at the ERM
        # (defaults to 0.2*innerRad, or set when calling function)
        # between these, the value is interpolated based on R,
        # and inside the value is interpolated between hInnerEdge and 0.2*innerEdge
        # if innerRad=0, then the mesh length is interpolated between
        # hEdge at the PM and 0.2*outerRad in the center
        # for one ellipsoid (innerRad = 0), if hEdge > 0.2*outerRad,
        # then lc = 0.2*outerRad in the whole volume
        # for two ellipsoids, if hEdge or hInnerEdge > 0.2*innerRad,
        # they are set to lc = 0.2*innerRad
        R_rel_outer = np.sqrt(
            (x / outerRad[0]) ** 2 + (y / outerRad[1]) ** 2 + (z / outerRad[2]) ** 2
        )
        if np.any(np.isclose(innerRad, 0)):
            lc3 = 0.2 * max(outerRad)
            innerRad_scale = 0
            in_outer = True
        else:
            R_rel_inner = np.sqrt(
                (x / innerRad[0]) ** 2 + (y / innerRad[1]) ** 2 + (z / innerRad[2]) ** 2
            )
            lc3 = 0.2 * max(innerRad)
            innerRad_scale = np.mean(
                [innerRad[0] / outerRad[0], innerRad[1] / outerRad[1], innerRad[2] / outerRad[2]]
            )
            in_outer = R_rel_inner > 1
        lc1 = hEdge
        lc2 = hInnerEdge
        if in_outer:
            lcTest = lc1 + (lc2 - lc1) * (1 - R_rel_outer) / (1 - innerRad_scale)
        else:
            lcTest = lc2 + (lc3 - lc2) * (1 - R_rel_inner)
        return min(lc3, lcTest)

    gmsh.model.mesh.setSizeCallback(meshSizeCallback)
    # set off the other options for mesh size determination
    gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)
    # this changes the algorithm from Frontal-Delaunay to Delaunay,
    # which may provide better results when there are larger gradients in mesh size
    gmsh.option.setNumber("Mesh.Algorithm", 5)

    gmsh.model.mesh.generate(3)
    rank = MPI.COMM_WORLD.rank
    tmp_folder = pathlib.Path(f"tmp_ellipsoid_{outerRad}_{innerRad}_{rank}")
    tmp_folder.mkdir(exist_ok=True)
    gmsh_file = tmp_folder / "ellipsoids.msh"
    gmsh.write(str(gmsh_file))
    gmsh.finalize()

    # return dolfin mesh of max dimension (parent mesh) and marker functions mf2 and mf3
    dmesh, mf2, mf3 = gmsh_to_dolfin(str(gmsh_file), tmp_folder, 3, comm)
    # remove tmp mesh and tmp folder
    gmsh_file.unlink(missing_ok=False)
    tmp_folder.rmdir()
    # return dolfin mesh, mf2 (2d tags) and mf3 (3d tags)
    return (dmesh, mf2, mf3)


def create_cylinders(
    outerRad: float = 1.0,
    innerRad: float = 0.0,
    outerLength: float = 10.0,
    innerLength: float = 8.0,
    hEdge: float = 0,
    hInnerEdge: float = 0,
    interface_marker: int = 12,
    outer_marker: int = 10,
    inner_vol_tag: int = 2,
    outer_vol_tag: int = 1,
    comm: MPI.Comm = d.MPI.comm_world,
    verbose: bool = False,
) -> Tuple[d.Mesh, d.MeshFunction, d.MeshFunction]:
    """
    Creates a mesh for use in examples that contains
    two distinct ellipsoid subvolumes with a shared interface
    surface. If the radius of the inner ellipsoid is 0, mesh a
    single ellipsoid.

    Args:
        outerRad: The radius of the outer cylinder
        innerRad: The radius of the inner cylinder
        outerLength: length of the outer cylinder
        innerLength: length of the inner cylinder
        hEdge: maximum mesh size at the outer edge
        hInnerEdge: maximum mesh size at the edge
            of the inner cylinder
        interface_marker: The value to mark facets on the interface with
        outer_marker: The value to mark facets on the outer ellipsoid with
        inner_vol_tag: The value to mark the inner spherical volume with
        outer_vol_tag: The value to mark the outer spherical volume with
        comm: MPI communicator to create the mesh with
        verbose: If true print gmsh output, else skip
    Returns:
        A triplet (mesh, facet_marker, cell_marker)
    """
    import gmsh

    if np.isclose(outerRad, 0):
        ValueError("Outer radius is equal to zero")
    if np.isclose(hEdge, 0):
        hEdge = 0.1 * max(outerRad)
    if np.isclose(hInnerEdge, 0):
        hInnerEdge = 0.2 * outerRad if np.isclose(innerRad, 0) else 0.2 * innerRad
    if innerRad > outerRad or innerLength >= outerLength:
        ValueError("Inner cylinder does not fit inside outer cylinder")
    # Create the two cylinder mesh using gmsh
    gmsh.initialize()
    gmsh.option.setNumber("General.Terminal", int(verbose))

    gmsh.model.add("twocylinders")
    # first add cylinder 1 of radius outerRad and center (0,0,0)
    outer_cylinder = gmsh.model.occ.addCylinder(0, 0, 0, 0, 0, outerLength, outerRad)
    if np.isclose(innerRad, 0):
        # Use outer_cylinder only
        gmsh.model.occ.synchronize()
        gmsh.model.add_physical_group(3, [outer_cylinder], tag=outer_vol_tag)
        facets = gmsh.model.getBoundary([(3, outer_cylinder)])
        gmsh.model.add_physical_group(2, [facets[0][1]], tag=outer_marker)
    else:
        # Add inner_cylinder (radius innerRad,
        # center (0,0,(outerLength-innerLength)/2))
        inner_cylinder = gmsh.model.occ.addCylinder(
            0, 0, (outerLength - innerLength) / 2, 0, 0, innerLength, innerRad
        )
        # Create interface between cylinders
        two_cylinders, (outer_cylinder_map, inner_cylinder_map) = gmsh.model.occ.fragment(
            [(3, outer_cylinder)], [(3, inner_cylinder)]
        )
        gmsh.model.occ.synchronize()

        # Get the outer boundary
        outer_shell = gmsh.model.getBoundary(two_cylinders, oriented=False)
        # Get the inner boundary
        inner_shell = gmsh.model.getBoundary(inner_cylinder_map, oriented=False)
        # Add physical markers for facets
        gmsh.model.add_physical_group(outer_shell[0][0], [outer_shell[0][1]], tag=outer_marker)
        gmsh.model.add_physical_group(inner_shell[0][0], [inner_shell[0][1]], tag=interface_marker)

        # Physical markers for
        all_volumes = [tag[1] for tag in outer_cylinder_map]
        inner_volume = [tag[1] for tag in inner_cylinder_map]
        outer_volume = []
        for vol in all_volumes:
            if vol not in inner_volume:
                outer_volume.append(vol)
        gmsh.model.add_physical_group(3, outer_volume, tag=outer_vol_tag)
        gmsh.model.add_physical_group(3, inner_volume, tag=inner_vol_tag)

    def meshSizeCallback(dim, tag, x, y, z, lc):
        # mesh length is hEdge at the PM (defaults to 0.1*outerRad,
        # or set when calling function) and hInnerEdge at the ERM
        # (defaults to 0.2*innerRad, or set when calling function)
        # between these, the value is interpolated based on r (polar coord),
        # and inside the value is interpolated between hInnerEdge and 0.2*innerEdge
        # if innerRad=0, then the mesh length is interpolated between
        # hEdge at the PM and 0.2*outerRad in the center
        # for one cylinder (innerRad = 0), if hEdge > 0.2*outerRad,
        # then lc = 0.2*outerRad in the whole volume
        # for two cylinders, if hEdge or hInnerEdge > 0.2*innerRad,
        # they are set to lc = 0.2*innerRad
        r_cur = np.sqrt(x**2 + y**2)
        if np.isclose(innerRad, 0):
            lc3 = 0.2 * outerRad
        else:
            lc3 = 0.2 * innerRad
        lc1 = hEdge
        lc2 = hInnerEdge
        if r_cur > innerRad:
            lcTest = lc1 + (lc2 - lc1) * (outerRad - r_cur) / (outerRad - innerRad)
        else:
            lcTest = lc2 + (lc3 - lc2) * (innerRad - r_cur) / innerRad
        return min(lc3, lcTest)

    gmsh.model.mesh.setSizeCallback(meshSizeCallback)
    # set off the other options for mesh size determination
    gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)
    # this changes the algorithm from Frontal-Delaunay to Delaunay,
    # which may provide better results when there are larger gradients in mesh size
    gmsh.option.setNumber("Mesh.Algorithm", 5)

    gmsh.model.mesh.generate(3)
    rank = MPI.COMM_WORLD.rank
    tmp_folder = pathlib.Path(f"tmp_cylinder_{outerRad}_{innerRad}_{rank}")
    tmp_folder.mkdir(exist_ok=True)
    gmsh_file = tmp_folder / "cylinders.msh"
    gmsh.write(str(gmsh_file))
    gmsh.finalize()

    # return dolfin mesh of max dimension (parent mesh) and marker functions mf2 and mf3
    dmesh, mf2, mf3 = gmsh_to_dolfin(str(gmsh_file), tmp_folder, 3, comm)
    # remove tmp mesh and tmp folder
    gmsh_file.unlink(missing_ok=False)
    tmp_folder.rmdir()
    # return dolfin mesh, mf2 (2d tags) and mf3 (3d tags)
    return (dmesh, mf2, mf3)


def create_ellipses(
    xrad_outer: float = 3.0,
    yrad_outer: float = 1.0,
    xrad_inner: float = 0.0,
    yrad_inner: float = 0.0,
    hEdge: float = 0.1,
    hInnerEdge: float = 0.1,
    interface_marker: int = 12,
    outer_marker: int = 10,
    inner_tag: int = 2,
    outer_tag: int = 1,
    comm: MPI.Comm = d.MPI.comm_world,
    verbose: bool = True,
) -> Tuple[d.Mesh, d.MeshFunction, d.MeshFunction]:
    """
    Creates a mesh for an ellipse surface,
    optionally includes an ellipse within the main ellipse
    Args:
        xrad_outer: outer radius assoc with x axis
        yrad_outer: outer radius assoc with y axis
        xrad_inner: x radius of inner ellipse (optional)
        yrad_inner: y radius of inner ellipse (optional)
        hEdge: mesh resolution at outer edge
        hInnerEdge: mesh resolution at edge of inner ellipse
        interface_marker: The value to mark facets on the interface with
        outer_marker: The value to mark facets on edge of the outer ellipse with
        inner_tag: The value to mark the inner ellipse surface with
        outer_tag: The value to mark the outer ellipse surface with
        comm: MPI communicator to create the mesh with
        verbose: If true print gmsh output, else skip
    Returns:
        Tuple (mesh, facet_marker (mf1), cell_marker(mf2))
    """
    import gmsh

    outerRad = [xrad_outer, yrad_outer]
    innerRad = [xrad_inner, yrad_inner]
    if np.any(np.isclose(outerRad, 0)):
        ValueError("One of the outer radii is equal to zero")
    if np.isclose(hEdge, 0):
        hEdge = 0.1 * max(outerRad)
    if np.isclose(hInnerEdge, 0):
        hInnerEdge = 0.2 * max(outerRad) if np.any(np.isclose(innerRad, 0)) else 0.2 * max(innerRad)
    if innerRad[0] > outerRad[0] or innerRad[1] > outerRad[1]:
        ValueError("Inner ellipse does not fit inside outer ellipse")
    # Create the two ellipse mesh using gmsh
    gmsh.initialize()
    gmsh.option.setNumber("General.Terminal", int(verbose))

    gmsh.model.add("ellipses")
    # first add ellipse 1 of radius outerRad and center (0,0,0)
    outer_ellipse = gmsh.model.occ.addDisk(0, 0, 0, xrad_outer, yrad_outer)
    if np.any(np.isclose(innerRad, 0)):
        # Use outer_ellipse only
        gmsh.model.occ.synchronize()
        gmsh.model.add_physical_group(2, [outer_ellipse], tag=outer_tag)
        facets = gmsh.model.getBoundary([(2, outer_ellipse)])
        assert len(facets) == 1
        gmsh.model.add_physical_group(1, [facets[0][1]], tag=outer_marker)
    else:
        # Add inner_ellipse (radius innerRad, center (0,0,0))
        inner_ellipse = gmsh.model.occ.addDisk(0, 0, 0, xrad_inner, yrad_inner)
        # Create interface between ellipses
        two_ellipses, (outer_ellipse_map, inner_ellipse_map) = gmsh.model.occ.fragment(
            [(2, outer_ellipse)], [(2, inner_ellipse)]
        )
        gmsh.model.occ.synchronize()

        # Get the outer boundary
        outer_shell = gmsh.model.getBoundary(two_ellipses, oriented=False)
        assert len(outer_shell) == 1
        # Get the inner boundary
        inner_shell = gmsh.model.getBoundary(inner_ellipse_map, oriented=False)
        assert len(inner_shell) == 1
        # Add physical markers for facets
        gmsh.model.add_physical_group(outer_shell[0][0], [outer_shell[0][1]], tag=outer_marker)
        gmsh.model.add_physical_group(inner_shell[0][0], [inner_shell[0][1]], tag=interface_marker)

        # Physical markers for
        all_surfs = [tag[1] for tag in outer_ellipse_map]
        inner_surf = [tag[1] for tag in inner_ellipse_map]
        outer_surf = []
        for surf in all_surfs:
            if surf not in inner_surf:
                outer_surf.append(surf)
        gmsh.model.add_physical_group(2, outer_surf, tag=outer_tag)
        gmsh.model.add_physical_group(2, inner_surf, tag=inner_tag)

    def meshSizeCallback(dim, tag, x, y, z, lc):
        # mesh length is hEdge at the PM (defaults to 0.1*outerRad,
        # or set when calling function) and hInnerEdge at the ERM
        # (defaults to 0.2*innerRad, or set when calling function)
        # between these, the value is interpolated based on R,
        # and inside the value is interpolated between hInnerEdge and 0.2*innerEdge
        # if innerRad=0, then the mesh length is interpolated between
        # hEdge at the PM and 0.2*outerRad in the center
        # for one ellipse (innerRad = 0), if hEdge > 0.2*outerRad,
        # then lc = 0.2*outerRad in the whole volume
        # for two ellipses, if hEdge or hInnerEdge > 0.2*innerRad,
        # they are set to lc = 0.2*innerRad
        R_rel_outer = np.sqrt((x / outerRad[0]) ** 2 + (y / outerRad[1]) ** 2)
        if np.any(np.isclose(innerRad, 0)):
            lc3 = 0.2 * max(outerRad)
            innerRad_scale = 0
            in_outer = True
        else:
            R_rel_inner = np.sqrt((x / innerRad[0]) ** 2 + (y / innerRad[1]) ** 2)
            lc3 = 0.2 * max(innerRad)
            innerRad_scale = np.mean([innerRad[0] / outerRad[0], innerRad[1] / outerRad[1]])
            in_outer = R_rel_inner > 1
        lc1 = hEdge
        lc2 = hInnerEdge
        if in_outer:
            lcTest = lc1 + (lc2 - lc1) * (1 - R_rel_outer) / (1 - innerRad_scale)
        else:
            lcTest = lc2 + (lc3 - lc2) * (1 - R_rel_inner)
        return min(lc3, lcTest)

    gmsh.model.mesh.setSizeCallback(meshSizeCallback)
    # set off the other options for mesh size determination
    gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
    gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)
    # this changes the algorithm from Frontal-Delaunay to Delaunay,
    # which may provide better results when there are larger gradients in mesh size
    gmsh.option.setNumber("Mesh.Algorithm", 5)

    gmsh.model.mesh.generate(2)
    rank = MPI.COMM_WORLD.rank
    tmp_folder = pathlib.Path(f"tmp_ellipse_{outerRad}_{innerRad}_{rank}")
    tmp_folder.mkdir(exist_ok=True)
    gmsh_file = tmp_folder / "ellipses.msh"
    gmsh.write(str(gmsh_file))
    gmsh.finalize()

    # return dolfin mesh of max dimension (parent mesh) and marker functions mf2 and mf3
    dmesh, mf1, mf2 = gmsh_to_dolfin(str(gmsh_file), tmp_folder, 2, comm)
    # remove tmp mesh and tmp folder
    gmsh_file.unlink(missing_ok=False)
    tmp_folder.rmdir()
    # return dolfin mesh, mf1 (1d tags) and mf2 (2d tags)
    return (dmesh, mf1, mf2)


def gmsh_to_dolfin(
    gmsh_file_name: str,
    tmp_folder: pathlib.Path = pathlib.Path("tmp_folder"),
    dimension: int = 3,
    comm: MPI.Comm = d.MPI.comm_world,
) -> Tuple[d.Mesh, d.MeshFunction, d.MeshFunction]:
    """
    Convert .msh file from gmsh to dolfin mesh
    and associated marker files (using meshio).
    Markers are assigned from gmsh mesh, any unassigned
    marker values are given value 0.
    Args:
        gmsh_file_name: .msh file (string)
        tmp_folder_name: folder name to store temporary mesh files
        dimension: dimension of parent mesh (int - either 2 or 3)
    Returns:
        Tuple containing:
            dMesh: Dolfin-style parent mesh
            mf_facet: markers for facets
            mf_cell: markers for cells
    """
    import meshio

    # load, convert to xdmf, and save as temp files
    mesh_in = meshio.read(gmsh_file_name)
    if dimension == 2:
        cell_type = "triangle"
        facet_type = "line"
    elif dimension == 3:
        cell_type = "tetra"
        facet_type = "triangle"
    else:
        ValueError(f"Mesh of dimension {dimension} not implemented")
    # convert cell mesh
    cells = mesh_in.get_cells_type(cell_type)
    cell_data = mesh_in.get_cell_data("gmsh:physical", cell_type)  # extract values of tags
    out_mesh_cell = meshio.Mesh(
        points=mesh_in.points,
        cells={cell_type: cells},
        cell_data={"mf_data": [cell_data]},
    )
    tmp_file_cell = tmp_folder / "tempmesh_cell.xdmf"
    meshio.write(tmp_file_cell, out_mesh_cell)
    # convert facet mesh
    facets = mesh_in.get_cells_type(facet_type)
    facet_data = mesh_in.get_cell_data("gmsh:physical", facet_type)  # extract values of tags
    out_mesh_facet = meshio.Mesh(
        points=mesh_in.points,
        cells={facet_type: facets},
        cell_data={"mf_data": [facet_data]},
    )
    tmp_file_facet = tmp_folder / "tempmesh_facet.xdmf"
    meshio.write(tmp_file_facet, out_mesh_facet)

    # convert xdmf mesh to dolfin-style mesh
    dmesh = d.Mesh(comm)
    mvc_cell = d.MeshValueCollection("size_t", dmesh, dimension)
    with d.XDMFFile(comm, str(tmp_file_cell)) as infile:
        infile.read(dmesh)
        infile.read(mvc_cell, "mf_data")
    mf_cell = d.cpp.mesh.MeshFunctionSizet(dmesh, mvc_cell)
    # set unassigned volumes to tag=0
    mf_cell.array()[np.where(mf_cell.array() > 1e9)[0]] = 0

    mvc_facet = d.MeshValueCollection("size_t", dmesh, dimension - 1)
    with d.XDMFFile(comm, str(tmp_file_facet)) as infile:
        infile.read(mvc_facet, "mf_data")
    mf_facet = d.cpp.mesh.MeshFunctionSizet(dmesh, mvc_facet)
    # set unassigned faces to tag=0
    mf_facet.array()[np.where(mf_facet.array() > 1e9)[0]] = 0

    # remove temp meshes
    tmp_file_cell.unlink(missing_ok=False)
    tmp_file_cell.with_suffix(".h5").unlink(missing_ok=False)
    tmp_file_facet.unlink(missing_ok=False)
    tmp_file_facet.with_suffix(".h5").unlink(missing_ok=False)
    # return dolfin mesh and mfs (marker functions)
    return (dmesh, mf_facet, mf_cell)


def write_mesh(
    mesh: d.Mesh,
    mf_facet: d.MeshFunction,
    mf_cell: d.MeshFunction,
    filename: pathlib.Path = pathlib.Path("DemoMesh.h5"),
):
    """
    Write 3D mesh, with cell markers (mf3)
    and facet markers (mf2) to hdf5 file and pvd files.
    """
    comm = mesh.mpi_comm()
    # extract dimensionality for meshfunctions
    cell_dim = mf_cell.dim()
    facet_dim = mf_facet.dim()
    # Write mesh and meshfunctions to file
    hdf5 = d.HDF5File(comm, str(filename.with_suffix(".h5")), "w")
    hdf5.write(mesh, "/mesh")
    hdf5.write(mf_cell, f"/mf{cell_dim}")
    hdf5.write(mf_facet, f"/mf{facet_dim}")
    # For visualization of domains
    (
        d.File(
            mesh.mpi_comm(),
            str(filename.with_stem(filename.stem + f"_mf{cell_dim}").with_suffix(".pvd")),
        )
        << mf_cell
    )
    (
        d.File(
            mesh.mpi_comm(),
            str(filename.with_stem(filename.stem + f"_mf{facet_dim}").with_suffix(".pvd")),
        )
        << mf_facet
    )
